#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype import hooks** (i.e., public-facing functions integrating high-level
:mod:`importlib` machinery required to implement :pep:`302`- and
:pep:`451`-compliant import hooks with the abstract syntax tree (AST)
transformations defined by the low-level :mod:`beartype.claw._ast.clawastmain`
submodule).

This private submodule is the main entry point for this subpackage. Nonetheless,
this private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ TODO                               }....................
#FIXME: Improve the beartype_package() and beartype_packages() functions to emit
#non-fatal warnings when the passed package or packages have already been
#imported (i.e., are in the "sys.modules" list).

# ....................{ IMPORTS                            }....................
from beartype.claw._package.clawpkgenum import BeartypeClawCoverage
from beartype.claw._package.clawpkgmain import hook_packages
from beartype.roar import BeartypeClawHookUnpackagedException
from beartype._cave._cavefast import CallableFrameType
from beartype._conf.confmain import BeartypeConf
from beartype._conf.confcommon import BEARTYPE_CONF_DEFAULT
from beartype._data.api.standard.datapy import SCRIPT_MODULE_NAME
from beartype._util.func.utilfuncfile import get_func_filename_or_none
from beartype._util.func.utilfuncframe import (
    get_frame,
    get_frame_module_name_or_none,
    get_frame_package_name_or_none,
)
from collections.abc import Iterable
from pathlib import PurePath

# ....................{ HOOKERS                            }....................
def beartype_all(
    # Optional keyword-only parameters.
    *,
    conf: BeartypeConf = BEARTYPE_CONF_DEFAULT,
) -> None:
    '''
    Register a new **universal beartype import path hook** (i.e., callable
    inserted to the front of the standard :mod:`sys.path_hooks` list recursively
    decorating *all* annotated callables, classes, and variable assignments
    across *all* submodules of *all* packages on the first importation of those
    submodules with the :func:`beartype.beartype` decorator, wrapping those
    callables and classes with performant runtime type-checking).

    This function is the runtime equivalent of a full-blown static type checker
    like ``mypy`` or ``pyright``, enabling full-stack runtime type-checking of
    the current app -- including submodules defined by both:

    * First-party proprietary packages directly authored for this app.
    * Third-party open-source packages authored and maintained elsewhere.

    This function is thread-safe.

    Usage
    -----
    This function is intended to be called from module scope as the first
    statement of the top-level ``__init__`` submodule of the top-level package
    of an app to be fully type-checked by :mod:`beartype`. This function then
    registers an import path hook type-checking *all* annotated callables,
    classes, and variable assignments across *all* submodules of *all* packages
    on the first importation of those submodules: e.g.,

    .. code-block:: python

       # At the very top of "muh_package.__init__":
       from beartype.claw import beartype_all
       beartype_all()  # <-- beartype all subsequent imports, yo

       # Import submodules *AFTER* calling beartype_all().
       from muh_package._some_module import muh_function  # <-- @beartype it!
       from yer_package.other_module import muh_class     # <-- @beartype it!

    Caveats
    -------
    **This function is not intended to be called from intermediary APIs,
    libraries, frameworks, or other middleware.** This function is *only*
    intended to be called from full stack end-user applications as a convenient
    alternative to manually passing the names of all packages to be type-checked
    to the more granular :func:`.beartype_packages` function. This function
    imposes runtime type-checking on downstream reverse dependencies that may
    not necessarily want, expect, or tolerate runtime type-checking. This
    function should typically *only* be called by proprietary packages not
    expected to be reused by others. Open-source packages are advised to call
    other functions instead.

    **tl;dr:** *Only call this function in non-reusable end-user apps.*

    Parameters
    ----------
    conf : BeartypeConf, optional
        **Beartype configuration** (i.e., dataclass configuring the
        :mod:`beartype.beartype` decorator for *all* decoratable objects
        recursively decorated by the path hook added by this function).
        Defaults to ``BeartypeConf()``, the default :math:`O(1)` configuration.

    Raises
    ------
    BeartypeClawHookException
        If the passed ``conf`` parameter is *not* a beartype configuration
        (i.e., :class:`BeartypeConf` instance).
    '''

    # The advantage of one-liners is the vantage of vanity.
    hook_packages(claw_coverage=BeartypeClawCoverage.PACKAGES_ALL, conf=conf)


def beartype_this_package(
    # Optional keyword-only parameters.
    *,
    conf: BeartypeConf = BEARTYPE_CONF_DEFAULT,
) -> None:
    '''
    Register a new **current package beartype import path hook** (i.e., callable
    inserted to the front of the standard :mod:`sys.path_hooks` list recursively
    applying the :func:`beartype.beartype` decorator to *all*
    annotated callables, classes, and variable assignments across *all*
    submodules of the current user-defined package calling this function on the
    first importation of those submodules).

    This function is thread-safe.

    Usage
    -----
    This function is intended to be called from module scope as the first
    statement of the top-level ``__init__`` submodule of any package to be
    type-checked by :mod:`beartype`. This function then registers an import path
    hook type-checking *all* annotated callables, classes, and variable
    assignments across *all* submodules of that package on the first importation
    of those submodules: e.g.,

    .. code-block:: python

       # At the very top of "muh_package.__init__":
       from beartype.claw import beartype_this_package
       beartype_this_package()  # <-- beartype all subsequent imports, yo

       # Import package submodules *AFTER* calling beartype_this_package().
       from muh_package._some_module import muh_function  # <-- @beartype it!
       from muh_package.other_module import muh_class     # <-- @beartype it!

    Parameters
    ----------
    conf : BeartypeConf, optional
        **Beartype configuration** (i.e., dataclass configuring the
        :mod:`beartype.beartype` decorator for *all* decoratable objects
        recursively decorated by the path hook added by this function).
        Defaults to ``BeartypeConf()``, the default :math:`O(1)` configuration.

    Raises
    ------
    BeartypeClawHookException
        If the passed ``conf`` parameter is *not* a beartype configuration
        (i.e., :class:`.BeartypeConf` instance).
    BeartypeClawHookUnpackagedException
        If this function is called from outside any package structure (e.g.,
        top-level module or executable script).
    '''

    # Stack frame encapsulating the user-defined lexical scope directly calling
    # this import hook.
    #
    # Note that:
    # * This call is guaranteed to succeed without error. Why? Because:
    #   * The current call stack *ALWAYS* contains at least one stack frame.
    #     Ergo, get_frame(0) *ALWAYS* succeeds without error.
    #   * The call to this import hook guaranteeably adds yet another stack
    #     frame to the current call stack. Ergo, get_frame(1) also *ALWAYS*
    #     succeeds without error in this context.
    # * This and the following logic *CANNOT* reasonably be isolated to a new
    #   private helper function. Why? Because this logic itself calls existing
    #   private helper functions assuming the caller to be at the expected
    #   position on the current call stack.
    frame_caller: CallableFrameType = get_frame(1)  # type: ignore[assignment,misc]

    # Fully-qualified name of the parent package of the child module defining
    # that caller if that module resides in some package *OR* the empty string
    # otherwise (i.e., if that module is a top-level module or script residing
    # outside any package structure).
    frame_package_name = get_frame_package_name_or_none(frame_caller)
    # print(f'beartype_this_package: {frame_caller_package_name}')
    # print(f'beartype_this_package: {repr(frame_caller)}')

    #FIXME: Is "pragma: no cover" accurate here? Is this condition untestable?
    # If that module has *NO* parent package, raise an exception. Why? Because
    # this function uselessly (but silently) reduces to a noop when called from
    # a top-level module or script residing outside any package. Why? Because
    # this function installs an import hook applicable only to subsequently
    # imported submodules of the current package. By definition, a top-level
    # module or script has *NO* package and thus *NO* sibling submodules and
    # thus *NO* meaningful imports to be hooked. To avoid unwanted confusion, we
    # intentionally notify the user with a loud exception.
    if not frame_package_name:  # pragma: no cover
        # Exception message to be raised below.
        exception_message: str = None  # type: ignore[assignment]

        # Fully-qualified name of the module encapsulating the caller.
        frame_module_name = get_frame_module_name_or_none(frame_caller)

        # If the caller is a script rather than a module, this name is the
        # useless magic string "__main__". In this case...
        if frame_module_name == SCRIPT_MODULE_NAME:
            # Absolute filename of this script if this script physically resides
            # on the local filesystem *OR* "None" otherwise (i.e., if this
            # script is dynamically defined in-memory).
            frame_filename = get_func_filename_or_none(frame_caller)

            # If this script physically exists...
            if frame_filename:
                # Prefix this message appropriately.
                exception_message_prefix = (
                    f'Top-level script "{frame_filename}" ')
            # Else, this script only exists in memory. In this case...
            else:
                # Prefix this message appropriately.
                exception_message_prefix = 'In-memory script '

                # Fabricate an arbitrary filename. Just do it!
                frame_filename = 'scripts/main.py'

            # Path object encapsulating this filename.
            frame_path = PurePath(frame_filename)

            # Basename of the parent directory containing this script, defined
            # as either...
            frame_package_basename = (
                # If this filename contains at least two basenames, then:
                # * The last basename is that of this script.
                # * The second-to-last basename is that of the parent directory
                #   containing this script.
                frame_path.parts[-2]
                if len(frame_path.parts) >= 2 else
                # Else, this filename contains only the basename of this script.
                # In this case, fabricate an arbitrary basename. Just do it!
                'scripts'
            )

            # Exception message to be raised below.
            exception_message = (
                f'{exception_message_prefix}resides outside package structure. '
                f'Consider calling another "beartype.claw" import hook. '
                f'However, note that only other modules will be type-checked. '
                f'"{frame_filename}" itself will remain unchecked. '
                f'All business logic should reside in submodules '
                f'subsequently imported by "{frame_filename}": e.g.,\n'
                f'    # Instead of this at the top of "{frame_filename}"...\n'
                f'    from beartype.claw import beartype_this_package  # <-- you are here\n'
                f'    beartype_this_package()                          # <-- feels bad\n'
                f'\n'
                f'    # ...pass the basename of the "{frame_package_basename}/" subdirectory explicitly.\n'
                f'    from beartype.claw import beartype_package  # <-- you want to be here\n'
                f'    beartype_package("{frame_package_basename}")  # <-- feels good\n'
                f'\n'
                f'    from {frame_package_basename}.main_submodule import main_func  # <-- still feels good\n'
                f'    main_func()                   # <-- *GOOD*! "beartype.claw" type-checks this\n'
                f'    some_global: str = 0xFEEDFACE  # <-- *BAD*! "beartype.claw" ignores this\n'
                f'This has been a message from your friendly neighbourhood bear.'
            )
        # Else, the caller is a module with a useful name. In this case, define
        # an exception message.
        #
        # Note that this edge case implies that this is a top-level module
        # residing outside a package that was *NOT* run as a script. Since this
        # should *BASICALLY* never occur, there isn't terribly much we can do.
        else:
            exception_message = (
                f'Top-level module "{frame_module_name}" '
                f'resides outside package structure but was '
                f'*NOT* directly run as a script. '
                f'"beartype.claw" import hooks require that modules either '
                f'reside inside a package structure or be '
                f'directly run as scripts. '
                f'Since neither applies here, you are now off the deep end. '
                f'@beartype no longer has any idea what is going on, sadly. '
                f'Consider directly decorating classes and functions by the '
                f'@beartype.beartype decorator instead: e.g.,\n'
                f'    # Instead of this at the top of "{frame_module_name}"...\n'
                f'    from beartype.claw import beartype_this_package  # <-- you are here\n'
                f'    beartype_this_package()                          # <-- feels bad\n'
                f'\n'
                f"    # ...go old-school like it's 2017 and you just don't care.\n"
                f'    from beartype import beartype  # <-- you want to be here\n'
                f'    @beartype  # <-- feels good, yet kinda icky at same time\n'
                f'    def spicy_func() -> str: ...  # <-- *GOOD*! @beartype type-checks this\n'
                f'    some_global: str = 0xFEEDFACE  # <-- *BAD*! @beartype ignores this, but what can you do\n'
                f'For your safety, @beartype will now crash and burn.'
            )

        # Raise an exception.
        raise BeartypeClawHookUnpackagedException(exception_message)
    # Else, that module has a parent package.

    # Add a new import path hook beartyping this package.
    hook_packages(
        claw_coverage=BeartypeClawCoverage.PACKAGES_ONE,
        package_name=frame_package_name,
        conf=conf,
    )


#FIXME: Add a "Usage" docstring section resembling that of the docstring for the
#beartype_this_package() function.
def beartype_package(
    # Mandatory parameters.
    package_name: str,

    # Optional keyword-only parameters.
    *,
    conf: BeartypeConf = BEARTYPE_CONF_DEFAULT,
) -> None:
    '''
    Register a new **single package beartype import path hook** (i.e., callable
    inserted to the front of the standard :mod:`sys.path_hooks` list recursively
    applying the :func:`beartype.beartype` decorator to *all* annotated
    callables, classes, and variable assignments across *all* submodules of the
    package with the passed names on the first importation of those submodules).

    This function is thread-safe.

    Parameters
    ----------
    package_name : str
        Fully-qualified name of the package to be type-checked.
    conf : BeartypeConf, optional
        **Beartype configuration** (i.e., dataclass configuring the
        :mod:`beartype.beartype` decorator for *all* decoratable objects
        recursively decorated by the path hook added by this function).
        Defaults to ``BeartypeConf()``, the default :math:`O(1)` configuration.

    Raises
    ------
    BeartypeClawHookException
        If either:

        * The passed ``conf`` parameter is *not* a beartype configuration (i.e.,
          :class:`BeartypeConf` instance).
        * The passed ``package_name`` parameter is either:

          * *Not* a string.
          * The empty string.
          * A non-empty string that is *not* a valid **package name** (i.e.,
            ``"."``-delimited concatenation of valid Python identifiers).
    '''

    # Add a new import path hook beartyping this package.
    hook_packages(
        claw_coverage=BeartypeClawCoverage.PACKAGES_ONE,
        package_name=package_name,
        conf=conf,
    )


#FIXME: Add a "Usage" docstring section resembling that of the docstring for the
#beartype_this_package() function.
def beartype_packages(
    # Mandatory parameters.
    package_names: Iterable[str],

    # Optional keyword-only parameters.
    *,
    conf: BeartypeConf = BEARTYPE_CONF_DEFAULT,
) -> None:
    '''
    Register a new **multiple package beartype import path hook** (i.e.,
    callable inserted to the front of the standard :mod:`sys.path_hooks` list
    recursively applying the :func:`beartype.beartype` decorator to *all*
    annotated callables, classes, and variable assignments across *all*
    submodules of all packages with the passed names on the first importation of
    those submodules).

    This function is thread-safe.

    Parameters
    ----------
    package_names : Iterable[str]
        Iterable of the fully-qualified names of one or more packages to be
        type-checked.
    conf : BeartypeConf, optional
        **Beartype configuration** (i.e., dataclass configuring the
        :mod:`beartype.beartype` decorator for *all* decoratable objects
        recursively decorated by the path hook added by this function).
        Defaults to ``BeartypeConf()``, the default :math:`O(1)` configuration.

    Raises
    ------
    BeartypeClawHookException
        If either:

        * The passed ``conf`` parameter is *not* a beartype configuration (i.e.,
          :class:`BeartypeConf` instance).
        * The passed ``package_names`` parameter is either:

          * Non-iterable (i.e., fails to satisfy the
            :class:`collections.abc.Iterable` protocol).
          * An empty iterable.
          * A non-empty iterable containing at least one item that is either:

            * *Not* a string.
            * The empty string.
            * A non-empty string that is *not* a valid **package name** (i.e.,
              ``"."``-delimited concatenation of valid Python identifiers).
    '''

    # Add a new import path hook beartyping these packages.
    hook_packages(
        claw_coverage=BeartypeClawCoverage.PACKAGES_MANY,
        package_names=package_names,
        conf=conf,
    )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`-, :pep:`585`-, and :pep:`646`-compliant **tuple type
hint utility** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.hint.pep.proposal.pep484585646` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ tester                     }....................
def test_is_hint_pep484585646_tuple_variadic() -> None:
    '''
    Test the
    :func:`beartype._util.hint.pep.proposal.pep484585646.is_hint_pep484585646_tuple_variadic`
    tester.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.hint.pep.proposal.pep484585646 import (
        is_hint_pep484585646_tuple_variadic)
    from beartype._util.hint.pep.proposal.pep646692 import (
        make_hint_pep646_tuple_unpacked_prefix)
    from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_11
    from typing import Tuple  # <-- intentionally import the PEP 484 variant

    # ....................{ PASS                           }....................
    # Assert this tester returns true for PEP 484-compliant variable-length
    # tuple type hints.
    assert is_hint_pep484585646_tuple_variadic(Tuple[int, ...])

    # Assert this tester returns true for PEP 585-compliant variable-length
    # tuple type hints.
    assert is_hint_pep484585646_tuple_variadic(tuple[str, ...])

    # If the active Python interpreter targets Python >= 3.11 and thus supports
    # PEP 646...
    if IS_PYTHON_AT_LEAST_3_11:
        # PEP 646-compliant variadic unpacked child tuple hint subscripted by
        # arbitrary child child hints.
        hint_pep646_tuple_unpacked_variadic = (
            make_hint_pep646_tuple_unpacked_prefix((str, ...)))

        # Assert this tester accepts this hint.
        assert is_hint_pep484585646_tuple_variadic(
            hint_pep646_tuple_unpacked_variadic) is True
    # Else, the active Python interpreter targets Python <= 3.10 and thus fails
    # to support PEP 646.

    # ....................{ FAIL                           }....................
    # Assert this tester returns false for arbitrary objects.
    assert is_hint_pep484585646_tuple_variadic(
        "And listen'd in sharp pain for Saturn's voice.") is False


def test_is_hint_pep484585646_tuple_empty() -> None:
    '''
    Test the
    :func:`beartype._util.hint.pep.proposal.pep484585646.is_hint_pep484585646_tuple_empty`
    tester.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.hint.pep.proposal.pep484585646 import (
        is_hint_pep484585646_tuple_empty)
    from typing import Tuple  # <-- intentionally import the PEP 484 variant

    # ....................{ PASS                           }....................
    # Assert this tester returns true for PEP 484-compliant empty tuple type
    # hints.
    assert is_hint_pep484585646_tuple_empty(Tuple[()]) is True

    # Assert this tester returns true for PEP 585-compliant empty tuple type
    # hints.
    assert is_hint_pep484585646_tuple_empty(tuple[()]) is True

    # ....................{ FAIL                           }....................
    # Assert this tester returns false for PEP 484-compliant non-empty tuple
    # type hints.
    assert is_hint_pep484585646_tuple_empty(Tuple[int, ...]) is False

    # Assert this tester returns false for PEP 585-compliant non-empty tuple
    # type hints.
    assert is_hint_pep484585646_tuple_empty(tuple[int, ...]) is False

    # Assert this tester returns false for arbitrary objects.
    assert is_hint_pep484585646_tuple_empty(
        "Groan'd for the old allegiance once more,") is False



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`589`-compliant **type alias reducers** (i.e., low-level
low-level callables converting higher-level :class:`typing.TypedDict` subclasses
to lower-level type hints more readily consumable by :mod:`beartype`).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._data.typing.datatypingport import Hint
from beartype._data.typing.datatyping import MappingStrToAny

# ....................{ REDUCERS                           }....................
#FIXME: Remove *AFTER* deeply type-checking typed dictionaries. For now,
#shallowly type-checking such hints by reduction to untyped dictionaries
#remains the sanest temporary work-around.
def reduce_hint_pep589(hint: Hint, exception_prefix: str) -> Hint:
    '''
    Reduce the passed :pep:`589`-compliant **typed dictionary** (i.e.,
    :class:`typing.TypedDict` subclass) to a lower-level type hint currently
    supported by :mod:`beartype`.

    This reducer is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as reducers cannot be memoized.

    Parameters
    ----------
    hint : Hint
        Typed dictionary to be reduced.
    exception_prefix : str
        Human-readable substring prefixing raised exception messages.

    Returns
    -------
    Hint
        Lower-level type hint currently supported by :mod:`beartype`.
    '''

    # Silently ignore all child type hints annotating this dictionary by
    # reducing this hint to the "Mapping" type hint. Yes, "Mapping" rather than
    # "dict". By PEP 589 edict:
    #     First, any TypedDict type is consistent with Mapping[str, object].
    return MappingStrToAny  # pyright: ignore



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype type-check call metadata dataclass** (i.e., class aggregating *all*
metadata required by the current call to the wrapper function type-checking a
:func:`beartype.beartype`-decorated callable).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import TYPE_CHECKING
from beartype._cave._cavemap import NoneTypeOr
from beartype._check.metadata.metadecor import BeartypeDecorMeta
from beartype._conf.confmain import BeartypeConf
from beartype._data.typing.datatyping import TypeStack
from beartype._data.typing.datatypingport import DictStrToHint
from collections.abc import Callable

# ....................{ CLASSES                            }....................
#FIXME: Unit test us up, please.
class BeartypeCheckMeta(object):
    '''
    **Beartype type-check call metadata** (i.e., object encapsulating *all*
    metadata required by each call to the wrapper function type-checking the
    callable currently being decorated by the :func:`beartype.beartype`
    decorator).

    Design
    ------
    This type-checking-time dataclass is effectively the proper subset of the
    comparable -- but *much* more complex in both space, time, and code
    complexity -- **decoration call metadata dataclass** (i.e.,
    :class:`beartype._check.metadata.metadecor.BeartypeDecorMeta`).
    Theoretically, this type-checking-time dataclass is thus redundant; the
    existing decoration call metadata dataclass could simply be used in lieu of
    this type-checking-time dataclass. Pragmatically, this type-checking-time
    dataclass significantly reduces the sheer quantity of metadata needed to
    type-check :func:`beartype.beartype`-decorated callables and thus the space
    consumption associated with that type-checking. In short, this is necessary.

    Attributes
    ----------
    cls_stack : TypeStack
        **Type stack** (i.e., either tuple of zero or more arbitrary types *or*
        :data:`None`). See also the parameter of the same name accepted by the
        :func:`beartype._decor.decorcore.beartype_object` function for details.
    conf : BeartypeConf
        **Beartype configuration** (i.e., self-caching dataclass encapsulating
        all flags, options, settings, and other metadata configuring the
        current decoration of the decorated callable).
    func : Callable
        **Decorated callable** (i.e., high-level callable currently being
        decorated by the :func:`beartype.beartype` decorator).
    func_annotations : dict[str, Hint]
        **Type hint dictionary** (i.e., mapping from the name of each annotated
        parameter accepted by the decorated callable to the type hint annotating
        that parameter).
    '''

    # ..................{ CLASS VARIABLES                    }..................
    # Slot all instance variables defined on this object to minimize the time
    # complexity of both reading and writing variables across frequently
    # called @beartype decorations. Slotting has been shown to reduce read and
    # write costs by approximately ~10%, which is non-trivial.
    __slots__ = (
        'cls_stack',
        'conf',
        'func',
        'func_annotations',
    )

    # Squelch false negatives from mypy. This is absurd. This is mypy. See:
    #     https://github.com/python/mypy/issues/5941
    if TYPE_CHECKING:
        cls_stack: TypeStack
        conf: BeartypeConf
        func: Callable
        func_annotations: DictStrToHint

    # Coerce instances of this class to be unhashable, preventing spurious
    # issues when accidentally passing these instances to memoized callables by
    # implicitly raising a "TypeError" exception on the first call to those
    # callables. There exists no tangible benefit to permitting these instances
    # to be hashed (and thus also cached), since these instances are:
    # * Specific to the decorated callable and thus *NOT* safely cacheable
    #   across functions applying to different decorated callables.
    #
    # See also:
    #     https://docs.python.org/3/reference/datamodel.html#object.__hash__
    __hash__ = None  # type: ignore[assignment]

    # ..................{ INITIALIZERS                       }..................
    def __init__(
        self,
        conf: BeartypeConf,
        cls_stack: TypeStack,
        func: Callable,
        func_annotations: DictStrToHint,
    ) -> None:
        '''
        Initialize this metadata with the passed parameters.

        Caveats
        -------
        **Avoid calling this low-level initializer directly.** Instead,
        instantiate instances of this dataclass by calling the
        :meth:`make_from_decor_meta` class method -- reducing existing
        instances of the parent :class:`.BeartypeDecorMeta` dataclass to
        instances of this child dataclass.

        Parameters
        ----------
        cls_stack : TypeStack
            **Type stack** (i.e., either tuple of zero or more arbitrary types
            *or* :data:`None`). See also the parameter of the same name accepted
            by the :func:`beartype._decor.decorcore.beartype_object` function
            for details.
        conf : BeartypeConf
            **Beartype configuration** (i.e., self-caching dataclass
            encapsulating all flags, options, settings, and other metadata
            configuring the current decoration of the decorated callable).
        func : Callable
            **Decorated callable** (i.e., high-level callable currently being
            decorated by the :func:`beartype.beartype` decorator).
        func_annotations : dict[str, Hint]
            **Type hint dictionary** (i.e., mapping from the name of each
            annotated parameter accepted by the decorated callable to the type
            hint annotating that parameter).
        '''
        assert isinstance(cls_stack, NoneTypeOr[tuple]), (
            f'{repr(cls_stack)} neither tuple nor "None".')
        assert isinstance(conf, BeartypeConf), (
            f'{repr(conf)} not beartype configuration.')
        assert callable(func), f'{repr(func)} uncallable.'
        assert isinstance(func_annotations, dict), (
            f'{repr(func_annotations)} not dictionary.')

        # Classify all passed parameters as instance variables.
        self.cls_stack = cls_stack
        self.conf = conf
        self.func = func
        self.func_annotations = func_annotations

    # ..................{ CLASS METHODS                      }..................
    @classmethod
    def make_from_decor_meta(
        cls, decor_meta: BeartypeDecorMeta) -> 'BeartypeCheckMeta':
        '''
        **Beartype type-check call metadata** (i.e., object encapsulating *all*
        metadata required by the current call to the wrapper function
        type-checking the callable currently being decorated by the
        :func:`beartype.beartype` decorator) reduced from the passed **beartype
        decorator call metadata** (i.e., object encapsulating *all* metadata for
        that callable).

        Parameters
        ----------
        decor_meta : BeartypeDecorMeta
            Beartype decorator call metadata to be reduced.
        '''
        assert isinstance(decor_meta, BeartypeDecorMeta)

        # Create and return a new instance of this child dataclass reduced from
        # the passed parent dataclass.
        return BeartypeCheckMeta(
            conf=decor_meta.conf,
            cls_stack=decor_meta.cls_stack,
            func=decor_meta.func_wrappee,
            func_annotations=decor_meta.func_annotations,
        )


    @classmethod
    def make_from_decor_meta_kwargs(cls, **kwargs) -> 'BeartypeCheckMeta':
        '''
        **Beartype type-check call metadata** (i.e., object encapsulating *all*
        metadata required by the current call to the wrapper function
        type-checking the callable currently being decorated by the
        :func:`beartype.beartype` decorator) reduced from the passed **beartype
        decorator call metadata keyword parameters** (i.e., keyword parameters
        to be passed to the :meth:`BeartypeDecorMeta.reinit` method).

        This factory method is a high-level convenience principally intended to
        be called from unit tests.

        Parameters
        ----------
        All passed keyword parameters are passed as is to the
        :meth:`BeartypeDecorMeta.reinit` method.
        '''

        # Beartype decorator call metadata with which to instantiate a new
        # instance of this dataclass.
        decor_meta = BeartypeDecorMeta()
        decor_meta.reinit(**kwargs)

        # Beartype type-checking call metadata reduced from this metadata.
        return cls.make_from_decor_meta(decor_meta)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype decorator :mod:`functools`-specific unit tests.

This submodule unit tests the :func:`beartype.beartype` decorator with respect
the standard :mod:`functools` module.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_decor_functools_lru_cache() -> None:
    '''
    Test the :func:`beartype.beartype` decorator on
    :func:`functools.lru_cache`-based **memoized callables** (i.e., pure-Python
    callables decorated by that standard decorator, which then creates and
    returns low-level C-based callable objects memoizing those callables).
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import (
        BeartypeCallHintParamViolation,
        # BeartypeDecorWrappeeException,
    )
    from functools import lru_cache
    from pytest import raises

    # ....................{ IDEAL                          }....................
    @lru_cache(maxsize=3)
    @beartype
    def increment_int_goodly(n: int) -> int:
        '''
        Arbitrary :func:`functools.lru_cache`-memoized callable decorated by
        :func:`beartype.beartype` in the ideal order.
        '''

        return n + 1


    # Assert that the ideal memoized callable when passed a valid parameter
    # returns the expected return.
    assert increment_int_goodly(42) == 43

    # Assert that the ideal memoized callable when passed an invalid parameter
    # raises the expected exception.
    with raises(BeartypeCallHintParamViolation):
        increment_int_goodly('The lone couch of his everlasting sleep:—')

    # ....................{ NON-IDEAL                      }....................
    @beartype
    @lru_cache(maxsize=3)
    def increment_int_badly(n: int) -> int:
        '''
        Arbitrary :func:`functools.lru_cache`-memoized callable decorated by
        :func:`beartype.beartype` in a non-ideal order.
        '''

        return n + 1

    # Assert that the non-ideal memoized callable when passed a valid parameter
    # returns the expected value.
    assert increment_int_badly(24) == 25

    # Assert that the non-ideal memoized callable when passed an invalid
    # parameter raises the expected exception.
    with raises(BeartypeCallHintParamViolation):
        increment_int_badly('Gentle, and brave, and generous,—no lorn bard')

    # ....................{ CLASSES                        }....................
    @beartype
    class IntIncrementer(object):
        '''
        Arbitrary :func:`beartype.beartype`-decorated class defining various
        :func:`functools.lru_cache`-memoized methods exercising edge cases.
        '''

        # ....................{ NON-IDEAL ~ static         }....................
        @staticmethod
        @lru_cache(maxsize=3)
        def increment_int_badly_statically(n: int) -> int:
            '''
            Arbitrary :func:`functools.lru_cache`-memoized static method
            implicitly decorated by :func:`beartype.beartype` in a non-ideal
            order.

            This static method exercises an edge cases, as:

            #. The :func:`.lru_cache` decorator creates and returns a
               **pseudo-callable object** (i.e., object callable only due to the
               type of that object defining the ``__call__()`` dunder method)
               rather than a pure-Python function.
            #. The builtin :class:`staticmethod` type then creates and returns a
               C-based descriptor object calling that pseudo-callable object.
            #. The :func:`beartype.beartype` decorator then:

               #. Unwraps that C-based descriptor object to access that
                  pseudo-callable object.
               #. Recreates that C-based descriptor object to instead wrap a
                  dynamically generated type-checking wrapper function calling
                  that pseudo-callable object.

            In short, madness.
            '''

            return n + 1

    # Assert that the non-ideal memoized callable when passed a valid parameter
    # returns the expected value.
    assert IntIncrementer.increment_int_badly_statically(42) == 43

    # Assert that the non-ideal memoized callable when passed an invalid
    # parameter raises the expected exception.
    with raises(BeartypeCallHintParamViolation):
        IntIncrementer.increment_int_badly_statically(
            'As if the ebbing air had but one wave;')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **type hint sign mapping** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._data.hint.sign.datahintsignmap` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_hint_sign_origin_isinstanceable_to_args_len_range() -> None:
    '''
    Test the
    :obj:`beartype._data.hint.sign.datahintsignmap.HINT_SIGN_ORIGIN_ISINSTANCEABLE_TO_ARGS_LEN_RANGE`
    dictionary global.

    This test guarantees conformance between this global and the corresponding
    type hint factories published by the standard :mod:`typing` module.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    import typing
    from beartype._data.hint.sign.datahintsigns import HintSignTuple
    from beartype._data.hint.sign.datahintsignmap import (
        HINT_SIGN_ORIGIN_ISINSTANCEABLE_TO_ARGS_LEN_RANGE)

    # ....................{ ASSERTS                        }....................
    # For each sign uniquely identifying an official type hint factory
    # subscriptable by a known number of child type hints and a "range" object
    # describing that number...
    for hint_sign, args_len_range in (
        HINT_SIGN_ORIGIN_ISINSTANCEABLE_TO_ARGS_LEN_RANGE.items()):
        # print(f'Inspecting {args_len_belief}-argument sign {args_sign}...')

        # If this is the sign uniquely identifying fixed-length tuple type hints
        # (e.g., "tuple[int, str]"), silently ignore the corresponding
        # "typing.Tuple" type hint factory and proceed to the next sign. Why?
        # Because that factory ambiguously supports both fixed- and variable-
        # length tuple type hints and thus defines:
        #     >>> from typing import Tuple
        #     >>> Tuple._nparams
        #     -1
        #
        # Needless to say, "-1" is an invalid number of arguments. *sigh*
        if hint_sign is HintSignTuple:
            return
        # Else, this is *NOT* the sign uniquely identifying fixed-length tuple
        # type hints.

        # Type hint factory published by the "typing" module with this name.
        hint_factory = getattr(typing, hint_sign.name)

        # Maximum number of arguments believed to be accepted by this factory
        # according to this dictionary.
        #
        # Note that the "stop" instance variable defined by "range" objects is
        # exclusive rather than inclusive. Substracting 1 from that yields the
        # inclusive maximum of this range.
        ARGS_LEN_MAX_BELIEF = args_len_range.stop - 1

        # Maximum number of arguments actually accepted by this factory.
        # Thankfully, this factory conveniently defines a private "_nparams"
        # instance variable providing this maximum number.
        ARGS_LEN_MAX_ACTUAL: int = hint_factory._nparams

        # Assert that these numbers of arguments correspond.
        assert ARGS_LEN_MAX_BELIEF == ARGS_LEN_MAX_ACTUAL, (
            f'Type hint factory "typing.{hint_sign.name}" accepts a maximum of '
            f'{ARGS_LEN_MAX_ACTUAL} child type hints but expected to accept a maximum of '
            f'{ARGS_LEN_MAX_BELIEF} child type hints.'
        )



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype utility fixed list pool unit tests.**

This submodule unit tests the public API of the private
:mod:`beartype._util.cache.pool.utilcachepoollistfixed` submodule.
'''

# ....................{ IMPORTS                           }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from pytest import raises

# ....................{ TESTS ~ pool                      }....................
def test_listfixed_pool_pass() -> None:
    '''
    Test successful usage of the
    :mod:`beartype._util.cache.pool.utilcachepoollistfixed` submodule.
    '''

    # Defer test-specific imports.
    from beartype._util.cache.pool.utilcachepoollistfixed import (
        acquire_fixed_list, release_fixed_list)

    # Acquire a fixed list of some length.
    moloch_solitude_filth = acquire_fixed_list(size=3)

    # Initialize this list.
    moloch_solitude_filth[:] = (
        'Moloch! Solitude! Filth! Ugliness! Ashcans and unobtainable dollars!',
        'Children screaming under the stairways! Boys sobbing in armies! Old',
        'men weeping in the parks!',
    )

    # Acquire another fixed list of the same length.
    moloch_whose = acquire_fixed_list(size=3)

    # Initialize this list.
    moloch_whose[:] = (
        'Moloch whose mind is pure machinery! Moloch whose blood is running',
        'money! Moloch whose fingers are ten armies! Moloch whose breast is a',
        'cannibal dynamo! Moloch whose ear is a smoking tomb!',
    )

    # Assert the contents of these lists to still be as expected.
    assert moloch_solitude_filth[0] == (
        'Moloch! Solitude! Filth! Ugliness! Ashcans and unobtainable dollars!')
    assert moloch_whose[-1] == (
        'cannibal dynamo! Moloch whose ear is a smoking tomb!')

    # Release these lists.
    release_fixed_list(moloch_solitude_filth)
    release_fixed_list(moloch_whose)


def test_listfixed_pool_fail() -> None:
    '''
    Test unsuccessful usage of the
    :mod:`beartype._util.cache.pool.utilcachepoollistfixed` submodule.
    '''

    # Defer test-specific imports.
    from beartype._util.cache.pool.utilcachepoollistfixed import (
        acquire_fixed_list)
    from beartype.roar._roarexc import _BeartypeUtilCachedFixedListException

    # Assert that fixed lists may only be acquired with positive integers.
    with raises(_BeartypeUtilCachedFixedListException):
        acquire_fixed_list((
            'Moloch the incomprehensible prison! Moloch the crossbone soulless',
            'jailhouse and Congress of sorrows! Moloch whose buildings are',
            'judgment! Moloch the vast stone of war! Moloch the stunned',
            'governments!',
        ))
    with raises(_BeartypeUtilCachedFixedListException):
        acquire_fixed_list(-67)

# ....................{ TESTS ~ type                      }....................
def test_listfixed_type_pass() -> None:
    '''
    Test successful usage of the
    :mod:`beartype._util.cache.pool.utilcachepoollistfixed.FixedList` type.
    '''

    # Defer test-specific imports.
    from beartype._util.cache.pool.utilcachepoollistfixed import FixedList

    # Fixed list to be tested.
    fixed_list = FixedList(size=11)

    # Assert this list to be of the expected size.
    assert len(fixed_list) == 11

    # Assert that list slicing to an iterable of the same length succeeds.
    fixed_list[:] = (
        'Love the quick profit, the annual raise,',
        'vacation with pay. Want more',
        'of everything ready-made. Be afraid',
        'to know your neighbors and to die.',
        'And you will have a window in your head.',
        'Not even your future will be a mystery',
        'any more. Your mind will be punched in a card',
        'and shut away in a little drawer.',
        'When they want you to buy something',
        'they will call you. When they want you',
        'to die for profit they will let you know.',
    )
    assert fixed_list[ 0] == 'Love the quick profit, the annual raise,'
    assert fixed_list[10] == 'to die for profit they will let you know.'

    # Assert that list copying succeeds.
    fixed_list_copy = fixed_list.copy()
    assert isinstance(fixed_list_copy, FixedList)
    assert fixed_list_copy == fixed_list


def test_listfixed_type_fail() -> None:
    '''
    Test unsuccessful usage of the
    :mod:`beartype._util.cache.pool.utilcachepoollistfixed.FixedList` type.
    '''

    # Defer test-specific imports.
    from beartype._util.cache.pool.utilcachepoollistfixed import FixedList
    from beartype.roar._roarexc import _BeartypeUtilCachedFixedListException

    # Fixed list to be tested.
    fixed_list = FixedList(size=8)
    fixed_list[:] = (
        'Ask the questions that have no answers.',
        'Invest in the millennium. Plant sequoias.',
        'Say that your main crop is the forest',
        'that you did not plant,',
        'that you will not live to harvest.',
        'Say that the leaves are harvested',
        'when they have rotted into the mold.',
        'Call that profit. Prophesy such returns.',
    )

    # Assert that fixed lists refuse to support item deletion.
    with raises(_BeartypeUtilCachedFixedListException):
        del fixed_list[0]

    # Assert that fixed lists refuse to support in-place addition.
    with raises(_BeartypeUtilCachedFixedListException):
        fixed_list += (
            'Put your faith in the two inches of humus',
            'that will build under the trees',
            'every thousand years.',
            'Listen to carrion – put your ear',
            'close, and hear the faint chattering',
            'of the songs that are to come.',
            'Expect the end of the world. Laugh.',
            'Laughter is immeasurable. Be joyful',
            'though you have considered all the facts.',
        )

    # Assert that fixed lists refuse to support in-place multiplication.
    with raises(_BeartypeUtilCachedFixedListException):
        fixed_list *= 0xDEADBEEF

    #FIXME: Disabled. For efficiency, fixed lists currently permissively
    #support slicing to an iterable of differing length. It is what it is.
    # # Assert that fixed lists refuse to support slicing to an iterable of
    # # differing length.
    # with raises(_BeartypeUtilCachedFixedListException):
    #     fixed_list[0:2] = (
    #         'Go with your love to the fields.',
    #         'Lie down in the shade. Rest your head',
    #         'in her lap. Swear allegiance',
    #         'to what is nighest your thoughts.',
    #         'As soon as the generals and the politicos',
    #         'can predict the motions of your mind,',
    #         'lose it. Leave it as a sign',
    #         'to mark the false trail, the way',
    #         'you didn’t go. Be like the fox',
    #         'who makes more tracks than necessary,',
    #         'some in the wrong direction.',
    #         'Practice resurrection.',
    #     )

    # Assert that fixed lists refuse to support appending.
    with raises(_BeartypeUtilCachedFixedListException):
        fixed_list.append('Love the world. Work for nothing.')

    # Assert that fixed lists refuse to support extending.
    with raises(_BeartypeUtilCachedFixedListException):
        fixed_list.extend((
            'Take all that you have and be poor.',
            'Love someone who does not deserve it.',
        ))

    # Assert that fixed lists refuse to support clearing.
    with raises(_BeartypeUtilCachedFixedListException):
        fixed_list.clear()

    # Assert that fixed lists refuse to support popping.
    with raises(_BeartypeUtilCachedFixedListException):
        fixed_list.pop()

    # Assert that fixed lists refuse to support removal.
    with raises(_BeartypeUtilCachedFixedListException):
        fixed_list.remove('Invest in the millennium. Plant sequoias.')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **Decidedly Object-Oriented Runtime-checking (DOOR) procedural
type-checkers** (i.e., high-level functions type-checking arbitrary objects
against type hints at *any* time during the lifecycle of the active Python
process).
'''

# ....................{ TODO                               }....................
#FIXME: Consider adding the following new tester:
#    def is_objects_similar(
#        # Mandatory parameters.
#        obj: object,
#        schema: object,
#
#        # Optional keyword-only parameters.
#        *,
#        conf: BeartypeConf = BEARTYPE_CONF_DEFAULT,
#    ) -> bool:
#        schema_hint = infer_hint(schema, conf=conf)
#        return is_bearable(obj, schema_hint, conf=conf)
#
#This test implements a new **typing structural similarity** metric answering
#the question: "Do these two objects have a similar structure?" This is a softer
#metric than object equality, which is a softer metric than object identity.
#Nonetheless, this metric is *INCREDIBLY* useful for certain use cases: e.g.,
#data validation, where one predefined exemplar archetypal object defined by the
#developer serves as a templating "schema" used to validate other arbitrary
#objects defined and passed in by users.

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# CAUTION: This submodule intentionally does *not* import the
# @beartype.beartype decorator. Why? Because that decorator conditionally
# reduces to a noop under certain contexts (e.g., `python3 -O` optimization),
# whereas the API defined by this submodule is expected to unconditionally
# operate as expected regardless of the current context.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype._check.checkmake import (
    make_func_raiser,
    make_func_tester,
)
from beartype._conf.confmain import BeartypeConf
from beartype._conf.confcommon import BEARTYPE_CONF_DEFAULT
from beartype._data.typing.datatypingport import (
    Hint,
    HintBare,
    TypeIs,
)
from beartype._data.typing.datatyping import T

# ....................{ VALIDATORS                         }....................
def die_if_unbearable(
    # Mandatory flexible parameters.
    obj: object,
    hint: Hint,

    # Optional keyword-only parameters.
    *,
    conf: BeartypeConf = BEARTYPE_CONF_DEFAULT,
    exception_prefix: str = 'die_if_unbearable() ',
) -> None:
    '''
    Raise an exception if the passed arbitrary object violates the passed type
    hint under the passed beartype configuration.

    To configure the type of violation exception raised by this function, set
    the :attr:`.BeartypeConf.violation_door_type` option of the passed ``conf``
    parameter accordingly.

    Parameters
    ----------
    obj : object
        Arbitrary object to be type-checked against this hint.
    hint : Hint
        Type hint to type-check this object against.
    conf : BeartypeConf, optional
        **Beartype configuration** (i.e., self-caching dataclass encapsulating
        all settings configuring type-checking for the passed object). Defaults
        to ``BeartypeConf()``, the default :math:`O(1)` constant-time
        configuration.
    exception_prefix : str, optional
        Human-readable label prefixing the representation of this object in the
        exception message. Defaults to a reasonably sensible string.

    Raises
    ------
    ``conf.violation_door_type``
        If this object violates this hint.
    beartype.roar.BeartypeDecorHintNonpepException
        If this hint is *not* PEP-compliant (i.e., complies with *no* Python
        Enhancement Proposals (PEPs) currently supported by :mod:`beartype`).
    beartype.roar.BeartypeDecorHintPepUnsupportedException
        If this hint is currently unsupported by :mod:`beartype`.

    Examples
    --------
    .. code-block:: pycon

       >>> from beartype.door import die_if_unbearable
       >>> die_if_unbearable(['And', 'what', 'rough', 'beast,'], list[str])
       >>> die_if_unbearable(['its', 'hour', 'come', 'round'], list[int])
       beartype.roar.BeartypeDoorHintViolation: Object ['its', 'hour', 'come',
       'round'] violates type hint list[int], as list index 0 item 'its' not
       instance of int.
    '''
    # conf._is_debug = True

    # Memoized low-level type-checking raiser function either raising an
    # exception or emitting a warning only if the passed object passed violates
    # the type hint passed to this high-level type-checking raiser function.
    #
    # Note that parameters are intentionally passed positionally for efficiency.
    # Since make_func_raiser() is memoized, passing parameters by keyword would
    # raise a non-fatal
    # "_BeartypeUtilCallableCachedKwargsWarning" warning.
    func_raiser = make_func_raiser(hint, conf, exception_prefix)  # pyright: ignore

    # Either raise an exception or emit a warning only if the passed object
    # violates this hint.
    func_raiser(obj)  # pyright: ignore

# ....................{ TESTERS ~ is_bearable              }....................
def is_bearable(
    # Mandatory flexible parameters.
    obj: object,
    hint: HintBare[T],

    # Optional keyword-only parameters.
    *,
    conf: BeartypeConf = BEARTYPE_CONF_DEFAULT,
) -> TypeIs[T]:  # pyright: ignore
    '''
    :data:`True` only if the passed arbitrary object satisfies the passed
    type hint under the passed beartype configuration.

    Note that this tester is a :pep:`647`- and :pep:`742`-compliant **smart type
    guard** (i.e., annotated by either the :pep:`742`-compliant return type hint
    ``typing.TypeIs[T]`` *or* the :pep:`647`-compliant return type hint
    ``typing.TypeGuard[T]`` conditionally depending on whether the active Python
    interpreter targets Python >= 3.13 or not). Specifically:

    * If the passed type hint is a valid **type** (i.e., subclass of the builtin
      :class:`type` superclass), this tester is a general-purpose type guard
      that performs type narrowing on the passed object: e.g.,

      .. code-block:: python

         from beartype.door import is_bearable

         def narrow_types_like_a_boss_with_beartype(lst: list[int | str]):
             # If "lst" is a list of integers, instruct static type-checkers
             # (e.g., mypy, pyright) that the call to munch_list_of_integers()
             # function expecting a list of integers is valid.
             #
             # Note that this works only because the is_bearable() tester
             # complies with PEP 647. If this were *NOT* the case, then static
             # type-checkers would raise a type-checking violation here.
             if is_bearable(lst, list[int]):
                 munch_list_of_integers(lst)
             # Else if "lst" is a list of strings, behave similarly as above.
             elif is_bearable(lst, list[str]):
                 munch_list_of_strings(lst)

         def munch_list_of_strings(lst: list[str]): ...
         def munch_list_of_integers(lst: list[int]): ....

    * If the passed type hint is *not* a valid type (e.g., the
      :pep:`586`-compliant ``typing.Literal['totally', 'not', 'a', 'type']``,
      which is clearly *not* a type), this tester is *not* a type guard and thus
      performs *no* type narrowing on the passed object. This is due to
      inadequacies in :pep:`647` rather than :mod:`beartype`.

    Parameters
    ----------
    obj : object
        Arbitrary object to be tested against this hint.
    hint : Hint
        Type hint to test this object against.
    conf : BeartypeConf, optional
        **Beartype configuration** (i.e., self-caching dataclass encapsulating
        all settings configuring type-checking for the passed object). Defaults
        to ``BeartypeConf()``, the default constant-time configuration.

    Returns
    -------
    bool
        :data:`True` only if this object satisfies this hint.

    Raises
    ------
    beartype.roar.BeartypeConfException
        If this configuration is *not* a :class:`BeartypeConf` instance.
    beartype.roar.BeartypeDecorHintForwardRefException
        If this hint contains one or more relative forward references, which
        this tester explicitly prohibits to improve both the efficiency and
        portability of calls to this tester.
    beartype.roar.BeartypeDecorHintNonpepException
        If this hint is *not* PEP-compliant (i.e., complies with *no* Python
        Enhancement Proposals (PEPs) currently supported by :mod:`beartype`).
    beartype.roar.BeartypeDecorHintPepUnsupportedException
        If this hint is currently unsupported by :mod:`beartype`.

    Examples
    --------
    .. code-block:: pycon

       >>> from beartype.door import is_bearable
       >>> is_bearable(['Things', 'fall', 'apart;'], list[str])
       True
       >>> is_bearable(['the', 'centre', 'cannot', 'hold;'], list[int])
       False
    '''

    # Memoized low-level type-checking tester function returning true only if
    # the object passed to that tester satisfies the type hint passed to this
    # high-level type-checking tester function.
    #
    # Note that parameters are intentionally passed positionally for efficiency.
    # Since make_func_tester() is memoized, passing parameters by keyword would
    # raise a non-fatal
    # "_BeartypeUtilCallableCachedKwargsWarning" warning.
    func_tester = make_func_tester(hint, conf)  # pyright: ignore

    # Return true only if the passed object satisfies this hint.
    return func_tester(obj)  # pyright: ignore

# ....................{ TESTERS                            }....................
def is_subhint(subhint: Hint, superhint: Hint) -> bool:
    '''
    :data:`True` only if the first passed hint is a **subhint** of the second
    passed hint, in which case this second hint is a **superhint** of this first
    hint.

    Equivalently, this tester returns :data:`True` only if *all* of the
    following conditions apply:

    * These two hints are **semantically related** (i.e., convey broadly similar
      semantics enabling these two hints to be reasonably compared). For
      example:

      * ``callable.abc.Iterable[str]`` and ``callable.abc.Sequence[int]`` are
        semantically related. These two hints both convey container semantics.
        Despite their differing child hints, these two hints are broadly similar
        enough to be reasonably comparable.
      * ``callable.abc.Iterable[str]`` and ``callable.abc.Callable[[], int]``
        are *not* semantically related. Whereas the first hint conveys a
        container semantic, the second hint conveys a callable semantic. Since
        these two semantics are unrelated, these two hints are dissimilar
        enough to *not* be reasonably comparable.

    * The first hint is **semantically equivalent** to or **narrower** than the
      second hint. Formally:

      * The first hint matches less than or equal to the total number of all
        possible objects matched by the second hint.
      * The size of the countably infinite set of all possible objects matched
        by the first hint is less than or equal to that of those matched by the
        second hint.

    * The first hint is **compatible** with the second hint. Since the first
      hint is semantically narrower than the second, APIs annotated by the first
      hint may safely replace that hint with the second hint; doing so preserves
      backward compatibility.

    Parameters
    ----------
    subhint : Hint
        Type hint or type to be tested as the subhint.
    superhint : Hint
        Type hint or type to be tested as the superhint.

    Returns
    -------
    bool
        :data:`True` only if this first hint is a subhint of this second hint.

    Examples
    --------
    .. code-block:: pycon

       >>> from beartype.door import is_subhint
       >>> is_subhint(int, int)
       True
       >>> is_subhint(Callable[[], list], Callable[..., Sequence[Any]])
       True
       >>> is_subhint(Callable[[], list], Callable[..., Sequence[int]])
       False
    '''

    # Avoid circular import dependencies.
    from beartype.door._cls.doorsuper import TypeHint

    # The one-liner is mightier than the... many-liner.
    return TypeHint(subhint).is_subhint(TypeHint(superhint))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype :pep:`544` **optimization layer** unit tests.

This submodule unit tests both the public *and* private API of the private
:mod:`beartype.typing._typingpep544` subpackage for sanity.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_typingpep544_metaclass() -> None:
    '''
    Test the private
    :class:`beartype.typing._typingpep544._CachingProtocolMeta` metaclass.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from abc import abstractmethod
    from beartype.typing import (
        Any,
        Protocol,
        TypeVar,
        runtime_checkable,
    )
    from beartype.typing._typingpep544 import _CachingProtocolMeta

    # ....................{ LOCALS                         }....................
    # Arbitrary type variable.
    _T_co = TypeVar('_T_co', covariant=True)

    # ....................{ CLASSES                        }....................
    # Can we really have it all?!
    @runtime_checkable  # <-- unnecessary at runtime, but Mypy is confused without it
    class SupportsRoundFromScratch(Protocol[_T_co]):
        __slots__: Any = ()
        @abstractmethod
        def __round__(self, ndigits: int = 0) -> _T_co:
            pass

    supports_round: SupportsRoundFromScratch = 0

    # ....................{ PASS                           }....................
    assert isinstance(supports_round, SupportsRoundFromScratch)
    assert issubclass(type(SupportsRoundFromScratch), _CachingProtocolMeta)


def test_typingpep544_superclass() -> None:
    '''
    Test the public :class:`beartype.typing.Protocol` superclass.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.typing import (
        Protocol as ProtocolFast,
        TypeVar,
    )
    from beartype.typing._typingpep544 import _ProtocolSlow as ProtocolSlow
    from pytest import raises

    # ....................{ LOCALS                         }....................
    # Arbitrary type variable.
    _T_co = TypeVar('_T_co', covariant=True)

    # Representations of protocols parametrized by one or more type variables.
    fast_repr = repr(ProtocolFast[_T_co])
    slow_repr = repr(ProtocolSlow[_T_co])

    # ....................{ PASS                           }....................
    # Assert that our caching protocol superclass memoizes subscriptions.
    assert ProtocolFast.__module__ == 'beartype.typing'
    assert ProtocolFast[_T_co] is ProtocolFast[_T_co]

    # Assert that the representation of a caching protocol is prefixed by the
    # expected prefix.
    assert fast_repr.startswith('beartype.typing.Protocol[')

    # Assert that these representation are suffixed by the same "["- and "]
    # "-delimited strings.
    fast_repr_suffix = fast_repr[fast_repr.rindex('['):]
    slow_repr_suffix = slow_repr[slow_repr.rindex('['):]
    assert fast_repr_suffix == slow_repr_suffix

    # ....................{ FAIL                           }....................
    # Assert that attempting to directly subscript the caching protocol
    # superclass by a non-type variable raises the expected exception.
    with raises(TypeError):
        ProtocolFast[str]


def test_typingpep544_subclass() -> None:
    '''
    Test expected behaviour of user-defined subclasses of the public
    :class:`beartype.typing.Protocol` superclass.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from abc import abstractmethod
    from beartype.typing import (
        AnyStr,
        Protocol,
        runtime_checkable
    )
    from pytest import raises

    # ....................{ CLASSES                        }....................
    class SupportsFeebleDreams(Protocol[AnyStr]):
        '''
        Arbitrary protocol directly subclassing the protocol superclass
        subscripted by one or more type variables, exercising subtle edge cases.
        '''

        @abstractmethod
        def torpor_of_the_year(self) -> AnyStr:
            pass


    class SupportsHiddenBuds(SupportsFeebleDreams[str]):
        '''
        Arbitrary protocol directly subclassing the above protocol superclass
        subscripted by one or more non-type variables satisfying the type
        variables subscripting that superclass, exercising subtle edge cases.
        '''

        @abstractmethod
        def dreamless_sleep(self) -> str:
            pass

    # ....................{ PASS                           }....................
    # Assert that optionally decorating protocols by the standard
    # @typing.runtime_checkable() decorator reduces to a noop.
    assert runtime_checkable(SupportsFeebleDreams) is SupportsFeebleDreams

    # Assert that a caching protocol subclass also memoizes subscriptions.
    assert SupportsFeebleDreams[str] is SupportsFeebleDreams[str]

    # Assert that optionally decorating abstract protocols by the standard
    # @typing.runtime_checkable() decorator reduces to a noop.
    assert runtime_checkable(SupportsFeebleDreams) is SupportsFeebleDreams

    # ....................{ FAIL                           }....................
    # Assert that attempting to decorate concrete protocol subclasses by
    # the same decorator raises the expected exception.
    with raises(TypeError):
        runtime_checkable(SupportsHiddenBuds)


def test_typingpep544_protocols_typing() -> None:
    '''
    Test the public retinue of ``beartype.typing.Supports*`` protocols with
    respect to both caching optimizations implemented by the public
    :class:`beartype.typing.Protocol` subclass *and* the private
    :class:`beartype.typing._typingpep544._CachingProtocolMeta` metaclass.
    '''

    # Defer test-specific imports.
    from decimal import Decimal
    from fractions import Fraction
    from beartype.typing import (
        SupportsAbs,
        SupportsBytes,
        SupportsComplex,
        SupportsFloat,
        SupportsIndex,
        SupportsInt,
        SupportsRound,
    )
    from beartype.typing._typingpep544 import _CachingProtocolMeta

    # Assert *ALL* standard protocols share the expected metaclass.
    for supports_t in (
        SupportsAbs,
        SupportsBytes,
        SupportsComplex,
        SupportsFloat,
        SupportsIndex,
        SupportsInt,
        SupportsRound,
    ):
        assert issubclass(type(supports_t), _CachingProtocolMeta)
        assert supports_t.__module__ == 'beartype.typing'

    def _assert_isinstance(*types: type, target_t: type) -> None:

        assert issubclass(
            target_t.__class__, _CachingProtocolMeta
        ), (
            f'{repr(target_t.__class__)} metaclass not '
            f'{repr(_CachingProtocolMeta)}.'
        )

        for t in types:
            v = t(0)
            assert isinstance(v, target_t), (
                f'{repr(t)} not instance of {repr(target_t)}.')


    # supports_abs: SupportsAbs = 0
    _assert_isinstance(
        int, float, bool, Decimal, Fraction, target_t=SupportsAbs)

    # supports_complex: SupportsComplex = Fraction(0, 1)
    _assert_isinstance(
        Decimal, Fraction, target_t=SupportsComplex)

    # supports_float: SupportsFloat = 0
    _assert_isinstance(
        int, float, bool, Decimal, Fraction, target_t=SupportsFloat)

    # supports_int: SupportsInt = 0
    _assert_isinstance(
        int, float, bool, target_t=SupportsInt)

    # supports_index: SupportsIndex = 0
    _assert_isinstance(
        int, bool, target_t=SupportsIndex)

    # supports_round: SupportsRound = 0
    _assert_isinstance(
        int, float, bool, Decimal, Fraction, target_t=SupportsRound)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype decorator call metadata dataclass** unit tests.

This submodule unit tests the :func:`beartype._check.metadata.metadecor`
submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_metadata_decor() -> None:
    '''
    Test the :func:`beartype._check.metadata.metadecor.BeartypeDecorMeta`
    dataclass.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import BeartypeConf
    from beartype.roar import BeartypeDecorWrappeeException
    from beartype._check.metadata.metadecor import BeartypeDecorMeta
    from beartype_test.a00_unit.data.data_type import function_lambda
    from pytest import raises

    # ....................{ LOCALS                         }....................
    # Arbitrary beartype decorator call metadata.
    bear_data = BeartypeDecorMeta()

    # ....................{ FAIL                           }....................
    # Assert this metadata to be unhashable.
    with raises(TypeError):
        hash(bear_data)

    # Assert that reinitializing this metadata an uncallable object to be
    # type-checked raises the expected exception.
    with raises(BeartypeDecorWrappeeException):
        bear_data.reinit(
            func='The fields, the lakes, the forests, and the streams,',
            conf=BeartypeConf(),
        )

    # Assert that reinitializing this metadata an uncallable object to be
    # unwrapped raises the expected exception.
    with raises(BeartypeDecorWrappeeException):
        bear_data.reinit(
            func=function_lambda,
            conf=BeartypeConf(),
            wrapper="Like serpents struggling in a vulture's grasp.",
        )

    # Assert that reinitializing this metadata with a C-based builtin function
    # raises the expected exception.
    with raises(BeartypeDecorWrappeeException):
        bear_data.reinit(func=iter, conf=BeartypeConf())

    # Assert that reinitializing this metadata with an invalid configuration
    # raises the expected exception.
    with raises(BeartypeDecorWrappeeException):
        bear_data.reinit(
            func=function_lambda,
            conf='Ocean, and all the living things that dwell',
        )

    # Assert that reinitializing this metadata with invalid class stacks raises
    # the expected exception.
    with raises(BeartypeDecorWrappeeException):
        bear_data.reinit(
            func=function_lambda,
            conf=BeartypeConf(),
            cls_stack="Shine in the rushing torrents' restless gleam,",
        )
    with raises(BeartypeDecorWrappeeException):
        bear_data.reinit(
            func=function_lambda,
            conf=BeartypeConf(),
            cls_stack=('Which from those secret chasms in tumult welling',),
        )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`- and :pep:`585`-compliant **generic type iterators**
(i.e., low-level callables generically iterating over both :pep:`484`- and
:pep:`585`-compliant generic class hierarchies).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep484585Exception
from beartype._check.convert.convmain import sanify_hint_child
from beartype._check.metadata.hint.hintsane import (
    HINT_SANE_IGNORABLE,
    HintSane,
)
from beartype._conf.confmain import BeartypeConf
from beartype._conf.confcommon import BEARTYPE_CONF_DEFAULT
from beartype._data.typing.datatyping import (
    HintSignOrNoneOrSentinel,
    TypeException,
    TypeStack,
)
from beartype._data.typing.datatypingport import Hint
from beartype._data.hint.sign.datahintsigncls import HintSign
from beartype._data.kind.datakindiota import SENTINEL
from beartype._util.cache.utilcachecall import callable_cached
from beartype._util.hint.pep.proposal.pep484585.generic.pep484585genget import (
    get_hint_pep484585_generic_base_extrinsic_sign_or_none,
    get_hint_pep484585_generic_bases_unerased,
)
from beartype._util.hint.pep.proposal.pep484585.generic.pep484585gentest import (
    is_hint_pep484585_generic_user)
from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign_or_none

# ....................{ PRIVATE ~ hints                    }....................
HintPep484585GenericUnsubbedBaseUnerased = tuple[HintSane, HintSign]
'''
:pep:`585`-compliant type hint matching each item of the tuple returned by the
:func:`.get_hint_pep484585_generic_unsubbed_bases_unerased` getter.
'''


HintPep484585GenericUnsubbedBasesUnerased = tuple[
    HintPep484585GenericUnsubbedBaseUnerased, ...]
'''
:pep:`585`-compliant type hint matching the tuple returned by the
:func:`.get_hint_pep484585_generic_unsubbed_bases_unerased` getter.
'''

# ....................{ GETTERS                            }....................
def get_hint_pep484585_generic_unsubbed_bases_unerased_kwargs(
    # Mandatory parameters.
    hint_sane: HintSane,

    # Optional parameters.
    cls_stack: TypeStack = None,
    conf: BeartypeConf = BEARTYPE_CONF_DEFAULT,
    exception_cls: TypeException = BeartypeDecorHintPep484585Exception,
    exception_prefix: str = '',
) -> HintPep484585GenericUnsubbedBasesUnerased:
    '''
    Unmemoized stub enabling callers to effectively pass keyword parameters to
    the memoized :func:`.get_hint_pep484585_generic_unsubbed_bases_unerased`
    getter, which accepts *only* positional parameters for efficiency.

    See Also
    --------
    :func:`.get_hint_pep484585_generic_unsubbed_bases_unerased`
        Further details.
    '''

    # Defer to this lower-level memoized getter.
    return get_hint_pep484585_generic_unsubbed_bases_unerased(
        hint_sane,
        cls_stack,
        conf,
        exception_prefix,
        exception_cls,
    )


#FIXME: Unit test us up, please.
#FIXME: Note that this would be, ideally, internally refactored to leverage the
#lower-level iter_hint_pep560_bases_unerased() iterator. We tried,
#actually... and failed hard. The current approach is "good enough." *shrug*
@callable_cached
def get_hint_pep484585_generic_unsubbed_bases_unerased(
    # Mandatory parameters.
    hint_sane: HintSane,

    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # CAUTION: Synchronize the order of these parameters with the
    # get_hint_pep484585_generic_unsubbed_bases_unerased_kwargs() wrapper above.
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # Optional parameters.
    #
    # Note that these parameters are intentionally ordered so as to trivialize
    # the calling convention. Since this getter is memoized, callers *MUST* pass
    # these parameters positionally rather than by keyword.
    cls_stack: TypeStack = None,
    conf: BeartypeConf = BEARTYPE_CONF_DEFAULT,
    exception_prefix: str = '',
    exception_cls: TypeException = BeartypeDecorHintPep484585Exception,
) -> HintPep484585GenericUnsubbedBasesUnerased:
    '''
    Tuple of the one or more **unerased pseudo-superclasses** (i.e., unignorable
    PEP-compliant type hints originally declared as transitive superclasses
    prior to type erasure) of the passed :pep:`484`- or :pep:`585`-compliant
    unsubscripted generic, effectively performing a breadth-first search (BFS)
    over these pseudo-superclasses.

    The tuple returned by this getter describes the full tree of all
    pseudo-superclasses by transitively visiting both all direct
    pseudo-superclasses of this generic *and* all indirect pseudo-superclasses
    transitively superclassing all direct pseudo-superclasses of this generic.
    For efficiency, this generator is internally implemented with an efficient
    imperative First In First Out (FILO) queue rather than an inefficient (and
    dangerous, due to both unavoidable stack exhaustion and avoidable infinite
    recursion) tree of recursive function calls.

    This getter is memoized for efficiency.

    Caveats
    -------
    **This generator exhibits** :math:`O(n)` **linear time complexity for**
    :math:`n` the number of transitive pseudo-superclasses of this generic. So,
    this generator is slow. The caller is expected to memoize *all* calls to
    this generator, which is itself *not* memoized.

    Design
    ------
    Note that there exist two kinds of pseudo-superclasses with respect to
    type-checking. Each pseudo-superclass yielded by this generator is either:

    * An **intrinsic pseudo-superclass** (i.e., whose type-checking is
      intrinsically defined as a type hint such that all data required to
      type-check this pseudo-superclass is fully defined by this hint). *All*
      intrinsic pseudo-superclasses are valid type hints. This is the common
      case and, indeed, almost all cases. Examples include :pep:`484`- and
      :pep:`585`-compliant subscripted container type hints: e.g.,

        .. code-block:: python

           # The PEP 585-compliant "list[T]" pseudo-superclass is a valid hint
           # whose type-checking is intrinsic to this hint.
           class GenericList[T](list[T]):
               def generic_method(self, arg: T) -> T:
                   return arg

    * An **extrinsic pseudo-superclass (i.e., whose type-checking is
      extrinsically defined by this unsubscripted generic such that only the
      combination of this pseudo-superclass and this unsubscripted generic
      suffices to provide all data required to type-check this
      pseudo-superclass). Extrinsic pseudo-superclasses are *not* necessarily
      valid type hints, though some might be. Examples include:

      * **Generic named tuples** (i.e., types subclassing both the
        :pep:`484`-compliant :class:`typing.Generic` superclass *and* the
        :pep:`484`-compliant :class:`typing.NamedTuple` superclass): e.g.,

        .. code-block:: python

           from typing import Generic, NamedTuple
           class GenericNamedTuple[T](NamedTuple, Generic[T]):
               generic_item: T

        When iterating over the :class:`typing.NamedTuple` pseudo-superclass of
        a generic typed dictionary, this generator yields the 2-tuple
        ``(hint_sane.hint, HintSignNamedTuple)`` (e.g.,
        ``(GenericNamedTuple, HintSignNamedTuple)`` for the above generic).

      * **Generic typed dictionaries** (i.e., types subclassing both the
        :pep:`484`-compliant :class:`typing.Generic` superclass *and* the
        :pep:`589`-compliant :class:`typing.TypedDict` superclass): e.g.,

        .. code-block:: python

           from typing import Generic, TypedDict
           class GenericTypedDict[T](TypedDict, Generic[T]):
               generic_item: T

        When iterating over the :class:`typing.TypedDict` pseudo-superclass of
        a generic typed dictionary, this generator yields the 2-tuple
        ``(hint_sane.hint, HintSignTypedDict)`` (e.g.,
        ``(GenericTypedDict, HintSignTypedDict)`` for the above generic).

    Motivation
    ----------
    Ideally, a BFS would *not* be necessary. Instead, pseudo-superclasses
    visited by this BFS should be visitable as is via whatever external parent
    BFS is currently iterating over the tree of all transitive type hints (e.g.,
    our code generation algorithm implemented by the
    :func:`beartype._check.code.codemain.make_func_pith_code` function).
    That's how we transitively visit all other kinds of type hints, right?
    Sadly, that simple solution fails to scale to all possible edge cases that
    arise with generics. Why? Because our code generation algorithm sensibly
    requires that *only* unignorable hints may be enqueued onto its outer BFS.
    Generics confound that constraint. Some pseudo-superclasses are
    paradoxically:

    * Ignorable from the perspective of code generation. *No* type-checking code
      should be generated for these pseudo-superclasses. See reasons below.
    * Unignorable from the perspective of algorithm visitation. These
      pseudo-superclasses generate *no* code but may themselves subclass other
      pseudo-superclasses for which type-checking code should be generated and
      which must thus be visited by our outer BFS.

    Paradoxical pseudo-superclasses include:

    * User-defined :pep:`484`-compliant subgenerics (i.e., user-defined generics
      subclassing one or more parent user-defined generic superclasses).
    * User-defined :pep:`544`-compliant subprotocols (i.e., user-defined
      protocols subclassing one or more parent user-defined protocol
      superclasses).

    Consider this example :pep:`544`-compliant subprotocol:

    .. code-block:: pycon

       >>> import typing as t
       >>> class UserProtocol(t.Protocol[t.AnyStr]): pass
       >>> class UserSubprotocol(UserProtocol[str], t.Protocol): pass
       >>> UserSubprotocol.__orig_bases__
       (UserProtocol[str], typing.Protocol)  # <-- good
       >>> UserProtocolUnerased = UserSubprotocol.__orig_bases__[0]
       >>> UserProtocolUnerased is UserProtocol
       False
       >>> isinstance(UserProtocolUnerased, type)
       False  # <-- bad

    :pep:`585`-compliant generics suffer no such issues:

    .. code-block:: pycon

       >>> from beartype._util.hint.pep.proposal.pep585 import is_hint_pep585_builtin_subbed
       >>> class UserGeneric(list[int]): pass
       >>> class UserSubgeneric(UserGeneric[int]): pass
       >>> UserSubgeneric.__orig_bases__
       (UserGeneric[int],)
       >>> UserGenericUnerased = UserSubgeneric.__orig_bases__[0]
       >>> isinstance(UserGenericUnerased, type)
       True  # <-- good
       >>> UserGenericUnerased.__mro__
       (UserGeneric, list, object)
       >>> is_hint_pep585_builtin_subbed(UserGenericUnerased)
       True

    Iteratively walking up the unerased inheritance hierarchy for any such
    paradoxical generic or protocol subclass (e.g., ``UserSubprotocol`` but
    *not* ``UserSubgeneric`` above) would visit a user-defined generic or
    protocol pseudo-superclass subscripted by type variables. Due to poorly
    defined obscurities in the :mod:`typing` implementation, that
    pseudo-superclass is *not* actually a class but rather an instance of a
    private :mod:`typing` class (e.g., :class:`typing._SpecialForm`). This
    algorithm would then detect that pseudo-superclass as neither a generic nor
    a :mod:`typing` object and thus raise an exception. Fortunately, that
    pseudo-superclass conveys no meaningful intrinsic semantics with respect to
    type-checking; its only use is to register its own pseudo-superclasses (one
    or more of which could convey meaningful intrinsic semantics with respect to
    type-checking) for visitation by this BFS.

    Parameters
    ----------
    hint_sane : HintSane
        **Sanified type hint metadata** (i.e., :data:`.HintSane` object)
        encapsulating the :pep:`484`- or :pep:`585`-compliant unsubscripted
        generic to be inspected.
    conf : BeartypeConf, optional
        **Beartype configuration** (i.e., self-caching dataclass encapsulating
        all settings configuring type-checking for the passed object). Defaults
        to :data:`.BEARTYPE_CONF_DEFAULT`, the default :math:`O(1)`
        type-checking configuration.
    cls_stack : TypeStack, optional
        **Type stack** (i.e., either a tuple of the one or more
        :func:`beartype.beartype`-decorated classes lexically containing the
        class variable or method annotated by this hint *or* :data:`None`).
        Defaults to :data:`None`.
    exception_cls : TypeException, default: BeartypeDecorHintPep484585Exception
        Type of exception to be raised in the event of fatal error. Defaults to
        :exc:`.BeartypeDecorHintPep484585Exception`.
    exception_prefix : str, default: ''
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Returns
    ------
    tuple[tuple[HintSane, HintSign], ...]
        2-tuple ``(hint_sane, hint_sign)``, where:

        * ``hint_sane`` is metadata encapsulating the sanification of an
          unignorable unerased transitive pseudo-superclass originally declared
          as a superclass prior to its type erasure of this unsubscripted
          generic.
        * ``hint_sign`` is the sign uniquely identifying this pseudo-superclass.
          Since this sign does *not* necessarily correspond to the sign returned
          by the :func:`.get_hint_pep_sign_or_none` getter when passed this
          pseudo-superclass, callers should take care to preserve this sign.

    Raises
    ------
    exception_cls
        If this hint is *not* a generic.

    See Also
    --------
    :func:`beartype._util.hint.pep.proposal.pep484585.generic.pep484585genget.get_hint_pep484585_generic_type_or_none`
        Further details.
    '''
    # assert is_hint_pep484585_generic_unsubbed(hint_sane.hint)
    assert isinstance(hint_sane, HintSane), (
        f'{repr(hint_sane)} not sanified metadata.')

    # ....................{ LOCALS                         }....................
    # This unsubscripted generic.
    hint = hint_sane.hint

    # List of 2-tuples "(hint_sane, hint_sign)" to be returned, where:
    # * "hint_sane" is metadata encapsulating the sanification of an unignorable
    #   unerased transitive pseudo-superclass originally declared as a
    #   superclass prior to its type erasure of this unsubscripted generic.
    # * "hint_sign" is the sign uniquely identifying this pseudo-superclass.
    hint_bases: list[HintPep484585GenericUnsubbedBaseUnerased] = []

    # Tuple of the one or more unerased pseudo-superclasses originally listed as
    # superclasses prior to their type erasure by this unsubscripted generic.
    hint_bases_direct = get_hint_pep484585_generic_bases_unerased(
        hint=hint,
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )
    # print(f'generic {hint} hint_bases_direct: {hint_bases_direct}')

    # Stack of unerased transitive pseudo-superclasses originally listed as
    # superclasses prior to their type erasure by this generic that have yet to
    # be visited by the breadth-first search (BFS) over these
    # pseudo-superclasses performed below. This stack is defined as a list of
    # 2-tuples "(hint_base, hint_base_subclass_sane)", where:
    # * "hint_base" is each such pseudo-superclass.
    # * "hint_base_subclass_sane" is the sanified hint metadata encapsulating
    #   the direct pseudo-*SUBCLASS* of this "hint_base", which will then be
    #   passed as the "hint_parent_sane" parameter to sanify this "hint_base".
    #   Specifically, this is either:
    #   * If this "hint_base" is a direct pseudo-superclass of the passed
    #     generic, the passed "hint_sane" metadata as is.
    #   * Else, this "hint_base" is a transitive pseudo-superclass two or more
    #     class levels above the passed generic. In this case, this is the
    #     "hint_base_sane" local calculated below.
    #
    # Note that this stack was previously defined as a fixed list. Sadly,
    # real-world torture tests like the third-party "redis.Redis" class of
    # "redis-py" fame obstruct the constraints imposed by fixed lists. That
    # class literally subclasses over 256 superclasses! For any fixed size,
    # there exists a real-world generic whose method-resolution order (MRO)
    # exceeds that size. The catastrophic cost of beartype raising spurious
    # exceptions against popular packages is significantly higher than any
    # negligible gains associated with microoptimizing this allocation away.
    hint_bases_visit: list[tuple[Hint, HintSane]] = []

    #FIXME: [SPEED] Optimize into a "while" loop. *sigh*
    # For the 0-based index of each direct pseudo-superclass of the passed
    # generic *AND* this direct pseudo-superclass...
    #
    # Note that this iteration guarantees this stack to initially be non-empty,
    # as "hint_bases_direct" itself is guaranteed to be non-empty.
    for hint_base_direct in hint_bases_direct:
        # Enqueue this direct pseudo-superclass and the sanified hint metadata
        # applicable to this pseudo-superclass onto this stack.
        hint_bases_visit.append((hint_base_direct, hint_sane))
    # print(f'generic pseudo-superclasses [initial]: {repr(hint_bases_direct)}')

    # ....................{ SEARCH                         }....................
    # While there exists at least one unerased transitive pseudo-superclasses to
    # be visited by this breadth-first search (BFS)...
    #
    # Each iteration of this search is subdivided into two phases, enabling this
    # search to discern between intrinsic and extrinsic pseudo-superclasses --
    # whose handling is fundamentally different. See the function docstring for
    # the distinction between the two.
    while hint_bases_visit:
        # ....................{ LOCALS                     }....................
        # Pop the next unerased transitive pseudo-superclass to be visited by
        # this BFS off this stack. Dismantled, this is:
        # * "hint_base", the currently visited transitive pseudo-superclass of
        #   the passed generic.
        # * "hint_base_subclass_sane", the sanified hint metadata applicable to
        #   this "hint_base". See "hint_bases_visit" above for details.
        hint_base, hint_base_subclass_sane = hint_bases_visit.pop()

        # ....................{ PHASE ~ extrinsic          }....................
        # In this first phase, we:
        # 1. Decide whether this pseudo-superclass is extrinsic.
        # 2. If so, return this subscripted generic (rather than this
        #    pseudo-superclass) *AND* the sign uniquely identifying this
        #    pseudo-superclass. To distinguish this sign from the normal sign
        #    identifying a hint, this sign is referred to as the "subsign"
        #    (i.e., subclass sign applicable to this subscripted generic
        #    subclassing this pseudo-superclass rather than this
        #    pseudo-superclass itself).
        # 3. Else, fallback to treating this pseudo-superclass as intrinsic by
        #    returning both this pseudo-superclass *AND* the sign uniquely
        #    identifying this pseudo-superclass. This is the common case.
        #
        # Note that:
        # * Extrinsic pseudo-superclasses are *EXTREMELY* rare. Almost all
        #   pseudo-superclasses are intrinsic.
        # * Extrinsic pseudo-superclasses are efficiently detectable in
        #   non-amortized O(1) time. Even though extrinsic pseudo-superclasses
        #   are rare, the cost of handling them is thankfully minimal.
        # * Extrinsic pseudo-superclasses *MUST* be detected before intrinsic
        #   pseudo-superclasses. Some extrinsic pseudo-superclasses (e.g.,
        #   "typing.TypedDict") are also valid type hints and thus also valid
        #   intrinsic pseudo-superclasses. Extrinsic pseudo-superclasses convey
        #   more fine-grained data for type-checking purposes than intrinsic
        #   pseudo-superclasses; the former are thus preferable to the latter.

        # Sign additionally identifying this pseudo-superclass if this
        # pseudo-superclass is extrinsic *OR* "None" otherwise (i.e., if this
        # pseudo-superclass is intrinsic).
        hint_base_extrinsic_sign = (
            get_hint_pep484585_generic_base_extrinsic_sign_or_none(
                hint_base=hint_base,
                exception_cls=exception_cls,
                exception_prefix=exception_prefix,
            ))

        # Hint to be sanified below, defaulting to this possibly insane
        # intrinsic pseudo-superclass.
        hint_sanify = hint_base

        # Sign with which to seed the first reduction of this hint to be
        # sanified, defaulting to the sentinel (effectively defaulting to the
        # standard sign uniquely identifying this hint returned by the
        # get_hint_pep_sign_or_none() getter).
        hint_sign_seed: HintSignOrNoneOrSentinel = SENTINEL

        # If this pseudo-superclass is extrinsic...
        if hint_base_extrinsic_sign is not None:
            # Re-sanify this *UNSUBSCRIPTED GENERIC* rather than this extrinsic
            # pseudo-superclass. Why? Because, by definition, an extrinsic
            # pseudo-superclass itself conveys insufficient metadata required to
            # type-check the unsubscripted generic subclassing that
            # pseudo-superclass; that metadata is extrinsic to that
            # pseudo-superclass, residing in this unsubscripted generic.
            hint_sanify = hint

            # Force the first reduction of this unsubscripted generic to be
            # performed against this extrinsic sign rather than
            # "HintSignPep484585GenericUnsubbed" (i.e., the intrinsic sign
            # identifying all unsubscripted generics by default).
            hint_sign_seed = hint_base_extrinsic_sign
        # Else, this pseudo-superclass is intrinsic. In this case, prefer the
        # defaults for these locals defaulted above.

        # Metadata encapsulating the sanification of this pseudo-superclass.
        hint_child_sane = sanify_hint_child(
            hint=hint_sanify,
            hint_parent_sane=hint_base_subclass_sane,
            hint_sign_seed=hint_sign_seed,
            cls_stack=cls_stack,
            conf=conf,
            exception_prefix=exception_prefix,
        )
        # print(f'generic {hint} base: {repr(hint_base)}')

        # If this pseudo-superclass is unignorable...
        if hint_child_sane is not HINT_SANE_IGNORABLE:
            # Sanified pseudo-superclass encapsulated by this metadata.
            hint_child = hint_child_sane.hint

            # If...
            if (
                # This pseudo-superclass is *NOT* actually just the passed
                # unsubscripted generic (i.e., the edge case that arises when
                # this pseudo-superclass is extrinsic) *AND*...
                hint_child is not hint and
                # This pseudo-superclass is itself a PEP 484- or 585-compliant
                # generic...
                is_hint_pep484585_generic_user(hint_child)
            ):
                # Then generate *NO* type-checking code for this
                # pseudo-superclass. Instead, only enqueue *ALL* parent
                # pseudo-superclasses of this child pseudo-superclass for
                # visitation by later iteration of this inner BFS.

                # Tuple of the one or more parent pseudo-superclasses of
                # this child pseudo-superclass.
                hint_child_bases = get_hint_pep484585_generic_bases_unerased(
                    hint=hint_child,
                    exception_cls=exception_cls,
                    exception_prefix=exception_prefix,
                )

                #FIXME: [SPEED] Optimize into a "while" loop, please. *sigh*
                # For each parent pseudo-superclass of this child
                # pseudo-superclass...
                for hint_child_base in hint_child_bases:
                    # print(f'hint_child_base: {hint_child_base}')
                    # print(f'hint_child_sane: {hint_child_sane}')
                    # print(f'hint_bases_visit: {len(hint_bases_visit)}')

                    # Enqueue this parent pseudo-superclass and the sanified
                    # hint metadata applicable to this pseudo-superclass onto
                    # this stack.
                    hint_bases_visit.append((hint_child_base, hint_child_sane))
            # Else, this pseudo-superclass is neither an ignorable user-defined
            # PEP 484-compliant generic *NOR* an ignorable 544-compliant
            # protocol. This implies this pseudo-superclass to be unignorable..
            else:
                # Sign uniquely identifying this pseudo-superclass if this
                # pseudo-superclass is PEP-compliant *OR* "None" otherwise
                # (i.e., if this pseudo-superclass is PEP-noncompliant), defined
                # as either...
                hint_child_sign = (
                    # If this pseudo-superclass is extrinsic *AND* the
                    # re-sanification of this unsubscripted generic as this
                    # extrinsic pseudo-superclass performed above preserved this
                    # unsubscripted generic as is (rather than reducing this
                    # unsubscripted generic to a lower-level hint), then this
                    # pseudo-superclass is still identified by this extrinsic
                    # sign. In this case, this extrinsic sign.
                    hint_base_extrinsic_sign
                    if (
                        hint_base_extrinsic_sign is not None and
                        hint_sanify is hint_child
                    ) else
                    # Else, either this pseudo-superclass is intrinsic *OR* the
                    # re-sanification of this unsubscripted generic as this
                    # extrinsic pseudo-superclass performed above reduced this
                    # unsubscripted generic to a lower-level hint. In either
                    # case, this pseudo-superclass is *NOT* identified by this
                    # extrinsic sign. In this case, default to inspecting the
                    # intrinsic sign uniquely identifying this child hint.
                    get_hint_pep_sign_or_none(hint_child)
                )

                # If this intrinsic pseudo-superclass is PEP-compliant, this
                # pseudo-superclass is a type hint conveying meaningful
                # semantics. In this case...
                if hint_child_sign is not None:
                    # print(f'Yielding generic {repr(hint)} base {repr(hint_child_sane)} ({hint_child_sign})...')

                    # Append the 2-tuple encapsulating both this
                    # pseudo-superclass and its identifying sign to this list to
                    # be returned, thus generating code type-checking this
                    # pseudo-superclass.
                    hint_bases.append((hint_child_sane, hint_child_sign))
                # Else, this pseudo-superclass is an isinstanceable type
                # conveying *NO* meaningful semantics and is thus effectively
                # ignorable. Why? Because the caller already type-checks this
                # pith against the generic subclassing this superclass and thus
                # this superclass as well inside an isinstance() call (e.g., in
                # the "CODE_PEP484585_GENERIC_PREFIX" snippet leveraged by the
                # "beartype._check.code.codemain" submodule).
        # Else, this pseudo-superclass is ignorable.
        # else:
        #     print(f'Ignoring generic {repr(hint)} base {repr(hint_base)}...')
        #     print(f'Is generic {hint} base {repr(hint_base)} type? {isinstance(hint_base, type)}')

    # ....................{ RETURN                         }....................
    # Return a tuple coerced from this list.
    return tuple(hint_bases)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`589`-compliant **type hint test data.**
'''

# ....................{ FIXTURES                           }....................
def hints_pep589_meta() -> 'List[HintPepMetadata]':
    '''
    Session-scoped fixture returning a list of :pep:`589`-compliant **type hint
    metadata** (i.e.,
    :class:`beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintPepMetadata`
    instances describing test-specific :pep:`589`-compliant sample type hints
    with metadata generically leveraged by various PEP-agnostic unit tests).
    '''

    # ..................{ IMPORTS                            }..................
    # Defer fixture-specific imports.
    from beartype.typing import (
        List,
        Type,
        Union,
    )
    from beartype._data.hint.sign.datahintsigns import (
        HintSignList,
        HintSignPep484585GenericSubbed,
        HintSignPep484585GenericUnsubbed,
        HintSignTypedDict,
    )
    from beartype._util.api.standard.utiltyping import get_typing_attrs
    from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_11
    from beartype_test.a00_unit.data.hint.util.data_hintmetacls import (
        HintPepMetadata,
        HintPithSatisfiedMetadata,
        HintPithUnsatisfiedMetadata,
    )

    from beartype_test.a00_unit.data.pep.data_pep484 import T

    # ..................{ LOCALS                             }..................
    # List of all PEP-specific type hint metadata to be returned.
    hints_pep_meta = []

    # ..................{ FACTORIES                          }..................
    # For each PEP-specific type hint factory importable from each currently
    # importable "typing" module...
    for TypedDict in get_typing_attrs('TypedDict'):
        # ..................{ SUBCLASSES                     }..................
        class ISeemAsInATranceSublimeAndStrange(TypedDict):
            '''
            Arbitrary empty typed dictionary annotated to require *no* key-value
            pairs.

            While patently absurd, this dictionary exercises an uncommon edge
            case in :pep:`589`.
            '''

            pass


        class DizzyRavine(ISeemAsInATranceSublimeAndStrange):
            '''
            Arbitrary non-empty typed dictionary annotated to require arbitrary
            key-value pairs, intentionally subclassing the empty typed
            dictionary subclass :class:`.ISeemAsInATranceSublimeAndStrange` to
            trivially exercise subclassability.
            '''

            # Arbitrary key whose value is annotated to be a PEP-noncompliant
            # instance of an isinstanceable type.
            and_when: str

            # Arbitrary key whose value is annotated to be a PEP-compliant union of
            # either a subclass of an issubclassable type or a PEP-noncompliant
            # instance of an isinstanceable type.
            I_gaze_on_thee: Union[bytes, Type[Exception]]


        #FIXME: Note that even this fails to suffice, thanks to *CRAY-CRAY*
        #subclassing logic that absolutely no one has ever exercised, but which
        #we'll nonetheless need to support. And I quoth:
        #    The totality flag only applies to items defined in the body of the
        #    TypedDict definition. Inherited items won't be affected, and
        #    instead use totality of the TypedDict type where they were defined.
        #    This makes it possible to have a combination of required and
        #    non-required keys in a single TypedDict type.
        #
        #Ergo, we need to additionally declare yet another new class subclassing
        #"ToMuse" but *NOT* explicitly subclassed with a "total" keyword
        #parameter. This clearly gets *EXTREMELY* ugly *EXTREMELY* fast, as
        #we'll now need to iterate over "hint.__mro__" in our code generation
        #algorithm. Well, I suppose we technically needed to do that anyway...
        #but still. Yikes!
        class ToMuse(TypedDict, total=False):
            '''
            Arbitrary non-empty typed dictionary annotated to require zero or
            more arbitrary key-value pairs.
            '''

            # Arbitrary key whose value is annotated to be a PEP-noncompliant
            # instance of an isinstanceable type.
            on_my_own: str

            # Arbitrary key whose value is annotated to be a PEP-compliant union
            # of either a subclass of an issubclassable type or a
            # PEP-noncompliant instance of an isinstanceable type.
            separate_fantasy: Union[Type[Exception], bytes]

        # ..................{ LISTS                          }..................
        # Add PEP-specific type hint metadata to this list.
        hints_pep_meta.extend((
            # ................{ TYPEDDICT                      }................
            # Empty typed dictionary. Look, this is ridiculous. What can you do?
            HintPepMetadata(
                hint=ISeemAsInATranceSublimeAndStrange,
                pep_sign=HintSignTypedDict,
                is_type_typing=False,
                piths_meta=(
                    # Empty dictionary instantiated with standard Python syntax.
                    HintPithSatisfiedMetadata({}),
                    # Empty dictionary instantiated from this typed dictionary.
                    HintPithSatisfiedMetadata(
                        ISeemAsInATranceSublimeAndStrange()),
                    # String constant.
                    HintPithUnsatisfiedMetadata(
                        pith='Hadithian bodies kindle Bodkin deathbeds',
                        # Match that the exception message raised for this
                        # object embeds the representation of the expected type.
                        exception_str_match_regexes=(r'\bMapping\b',),
                    ),
                    #FIXME: Uncomment *AFTER* deeply type-checking "TypedDict".
                    # # Non-empty dictionary.
                    # HintPithSatisfiedMetadata({
                    #     'Corinthian bodachean kinslayers lay': (
                    #         'wedded weal‐kith with in‐'),
                    # }),
                ),
            ),

            # Non-empty totalizing typed dictionary.
            HintPepMetadata(
                hint=DizzyRavine,
                pep_sign=HintSignTypedDict,
                is_type_typing=False,
                piths_meta=(
                    # Non-empty dictionary of the expected keys and values.
                    HintPithSatisfiedMetadata({
                        'and_when': 'Corrigible‐ragged gun corruptions within',
                        'I_gaze_on_thee': b"Hatross-ev-olved eleven imp's",
                    }),
                    # Non-empty dictionary of the expected keys and values
                    # instantiated from this typed dictionary.
                    HintPithSatisfiedMetadata(DizzyRavine(
                        and_when=(
                            'Machiavellian‐costumed, tumid stock fonts of a'),
                        I_gaze_on_thee=RuntimeError,
                    )),
                    # String constant.
                    HintPithUnsatisfiedMetadata(
                        pith='Matross‐elevated elven velvet atrocities of',
                        # Match that the exception message raised for this
                        # object embeds the representation of the expected type.
                        exception_str_match_regexes=(r'\bMapping\b',),
                    ),
                    # #FIXME: Uncomment *AFTER* deeply type-checking "TypedDict".
                    # # Empty dictionary.
                    # HintPithUnsatisfiedMetadata(
                    #     pith={},
                    #     # Match that the exception message raised for this object
                    #     # embeds the expected number of key-value pairs.
                    #     exception_str_match_regexes=(r'\b2\b',),
                    # ),
                    # # Non-empty dictionary of the expected keys but *NOT* values.
                    # HintPithUnsatisfiedMetadata(
                    #     pith={
                    #         'and_when': 'Matricidally',
                    #         'I_gaze_on_thee': (
                    #             'Hatchet‐cachepotting, '
                    #             'Scossetting mock misrule by'
                    #         ),
                    #     },
                    #     # Match that the exception message raised for this object
                    #     # embeds:
                    #     # * The name of the unsatisfied key.
                    #     # * The expected types of this key's value.
                    #     exception_str_match_regexes=(
                    #         r'\bI_gaze_on_thee\b',
                    #         r'\bbytes\b',
                    #     ),
                    # ),
                ),
            ),

            # Non-empty non-totalizing typed dictionary.
            HintPepMetadata(
                hint=ToMuse,
                pep_sign=HintSignTypedDict,
                is_type_typing=False,
                piths_meta=(
                    # Empty dictionary.
                    HintPithSatisfiedMetadata({}),
                    # Non-empty dictionary defining only one of the expected
                    # keys.
                    HintPithSatisfiedMetadata({
                        'on_my_own': (
                            'Spurned Court‐upturned, upper gladness, '
                            'edifyingly humidifying'),
                    }),
                    # Non-empty dictionary defining *ALL* of the expected keys,
                    # instantiated from this typed dictionary.
                    HintPithSatisfiedMetadata(ToMuse(
                        on_my_own=(
                            'Sepulchral epic‐âpostatizing home tombs metem‐'),
                        separate_fantasy=b'Macroglia relics',
                    )),
                    # String constant.
                    HintPithUnsatisfiedMetadata(
                        pith=(
                            'Psychotically tempered Into temporal '
                            'afterwork‐met portals portending a'
                        ),
                        # Match that the exception message raised for this
                        # object embeds the representation of the expected type.
                        exception_str_match_regexes=(r'\bMapping\b',),
                    ),
                    # #FIXME: Uncomment *AFTER* deeply type-checking "TypedDict".
                    # # Non-empty dictionary of the expected keys but *NOT* values.
                    # HintPithUnsatisfiedMetadata(
                    #     pith={
                    #         'on_my_own': (
                    #             'Psyche’s Maidenly‐enladened, '
                    #             'aidful Lads‐lickspittling Potenc‐ies —',
                    #         ),
                    #         'separate_fantasy': (
                    #             'Psychedelic metal‐metastasized, glib'),
                    #     },
                    #     # Match that the exception message raised for this object
                    #     # embeds:
                    #     # * The name of the unsatisfied key.
                    #     # * The expected types of this key's value.
                    #     exception_str_match_regexes=(
                    #         r'\bseparate_fantasy\b',
                    #         r'\bbytes\b',
                    #     ),
                    # ),
                ),
            ),

            # ................{ LISTS ~ nested                 }................
            # List of non-empty totalizing typed dictionaries.
            HintPepMetadata(
                hint=List[DizzyRavine],
                pep_sign=HintSignList,
                isinstanceable_type=list,
                piths_meta=(
                    # List of dictionaries of the expected keys and values.
                    HintPithSatisfiedMetadata([
                        {
                            'and_when': (
                                'Matriculating ‘over‐sized’ '
                                'research urchin Haunts of',
                            ),
                            'I_gaze_on_thee': b"Stands - to",
                        },
                        {
                            'and_when': (
                                'That resurrected, Erectile reptile’s '
                                'pituitary capitulations to',
                            ),
                            'I_gaze_on_thee': b"Strand our under-",
                        },
                    ]),
                    # List of string constants.
                    HintPithUnsatisfiedMetadata(
                        pith=[
                            'D-as K-apital, '
                            'notwithstanding Standard adiós‐',
                        ],
                        # Match that the exception message raised for this
                        # object embeds the representation of the expected type.
                        exception_str_match_regexes=(r'\bMapping\b',),
                    ),
                    # #FIXME: Uncomment *AFTER* deeply type-checking "TypedDict".
                    # # List of empty dictionaries.
                    # HintPithUnsatisfiedMetadata(
                    #     pith=[{}, {},],
                    #     # Match that the exception message raised for this object
                    #     # embeds the expected number of key-value pairs.
                    #     exception_str_match_regexes=(r'\b2\b',),
                    # ),
                    # # List of non-empty dictionaries, only one of which fails to
                    # # define both the expected keys and values.
                    # HintPithUnsatisfiedMetadata(
                    #     pith=[
                    #         {
                    #             'and_when': (
                    #                 'Diased capitalization of (or into)'),
                    #             'I_gaze_on_thee': (
                    #                 b'Witheringly dithering, dill husks of'),
                    #         },
                    #         {
                    #             'and_when': (
                    #                 'Will, like Whitewash‐ed, musky'),
                    #             'I_gaze_on_thee': 'Likenesses injecting',
                    #         },
                    #     ],
                    #     # Match that the exception message raised for this object
                    #     # embeds:
                    #     # * The index of the unsatisfied dictionary.
                    #     # * The name of the unsatisfied key.
                    #     # * The expected types of this key's value.
                    #     exception_str_match_regexes=(
                    #         r'\b1\b',
                    #         r'\bI_gaze_on_thee\b',
                    #         r'\bbytes\b',
                    #     ),
                    # ),
                ),
            ),
        ))

        # ..................{ VERSION                        }..................
        # If the active Python interpreter targets Python >= 3.11 and thus
        # supports PEP 484- and 589-compliant typed dictionary generics...
        if IS_PYTHON_AT_LEAST_3_11:
            # Defer version-specific imports.
            from beartype_test.a00_unit.data.pep.generic.data_pep589generic import (
                Pep589484TypedDictT)

            # Add version-specific type hint metadata to this list.
            hints_pep_meta.extend((
                # ................{ GENERICS                   }................
                # Non-empty totalizing unsubscripted typed dictionary generic
                # parametrized by one type variable.
                HintPepMetadata(
                    hint=Pep589484TypedDictT,
                    pep_sign=HintSignPep484585GenericUnsubbed,
                    generic_type=Pep589484TypedDictT,
                    is_type_typing=False,
                    typeargs_packed=(T,),
                    piths_meta=(
                        # Non-empty generic instance containing the expected keys
                        # and values.
                        HintPithSatisfiedMetadata(Pep589484TypedDictT(
                            key='Arches, and domes, and fiery galleries;')),
                        # String constant.
                        HintPithUnsatisfiedMetadata(
                            pith='And all its curtains of Aurorian clouds',
                            # Match that the exception message raised for this
                            # object embeds the representation of the expected type.
                            exception_str_match_regexes=(
                                r'\bPep589484TypedDictT\b',),
                        ),
                        # #FIXME: Uncomment *AFTER* deeply type-checking "TypedDict".
                        # # Empty generic instance.
                        # HintPithSatisfiedMetadata(Pep589484TypedDictT()),
                        # # Non-empty generic instance *NOT* containing the expected
                        # # keys.
                        # HintPithUnsatisfiedMetadata(
                        #     pith=Pep589484TypedDictT({
                        #         'and_when': 'Matricidally',
                        #         'I_gaze_on_thee': (
                        #             'Hatchet‐cachepotting, '
                        #             'Scossetting mock misrule by'
                        #         ),
                        #     }),
                        #     # Match that the exception message raised for this object
                        #     # embeds:
                        #     # * The name of the unsatisfied key.
                        #     # * The expected types of this key's value.
                        #     exception_str_match_regexes=(
                        #         r'\bI_gaze_on_thee\b',
                        #         r'\bbytes\b',
                        #     ),
                        # ),
                    ),
                ),

                # Non-empty totalizing subscripted typed dictionary generic
                # parametrized by one type variable.
                HintPepMetadata(
                    hint=Pep589484TypedDictT[str],
                    pep_sign=HintSignPep484585GenericSubbed,
                    generic_type=Pep589484TypedDictT,
                    is_type_typing=True,
                    is_typing=False,
                    piths_meta=(
                        # Non-empty generic instance containing the expected keys
                        # and values.
                        HintPithSatisfiedMetadata(Pep589484TypedDictT(
                            key="Flush'd angerly: while sometimes eagle's wings,")),
                        # String constant.
                        HintPithUnsatisfiedMetadata(
                            pith='Unseen before by Gods or wondering men,',
                            # Match that the exception message raised for this
                            # object embeds the representation of the expected type.
                            exception_str_match_regexes=(
                                r'\bPep589484TypedDictT\b',),
                        ),
                        # #FIXME: Uncomment *AFTER* deeply type-checking "TypedDict".
                        # # Empty generic instance.
                        # HintPithSatisfiedMetadata(Pep589484TypedDictT()),
                        # # Non-empty generic instance *NOT* containing the expected
                        # # keys.
                        # HintPithUnsatisfiedMetadata(
                        #     pith=Pep589484TypedDictT({
                        #         'and_when': 'Matricidally',
                        #         'I_gaze_on_thee': (
                        #             'Hatchet‐cachepotting, '
                        #             'Scossetting mock misrule by'
                        #         ),
                        #     }),
                        #     # Match that the exception message raised for this object
                        #     # embeds:
                        #     # * The name of the unsatisfied key.
                        #     # * The expected types of this key's value.
                        #     exception_str_match_regexes=(
                        #         r'\bI_gaze_on_thee\b',
                        #         r'\bbytes\b',
                        #     ),
                        # ),
                    ),
                ),
            ))

        # Else, the active Python interpreter targets Python <= 3.10 and thus
        # fails to support PEP 484- and 589-compliant typed dictionary generics.

    # ..................{ RETURN                             }..................
    # Return this list of all PEP-specific type hint metadata.
    return hints_pep_meta



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype decorator code generator.**

This private submodule dynamically generates both the signature and body of the
wrapper function type-checking all annotated parameters and return value of the
the callable currently being decorated by the :func:`beartype.beartype`
decorator in a general-purpose manner. For genericity, this relatively
high-level submodule implements *no* support for annotation-based PEPs (e.g.,
:pep:`484`); other lower-level submodules do so instead.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ TODO                               }....................
# All "FIXME:" comments for this submodule reside in this package's "__init__"
# submodule to improve maintainability and readability here.

# ....................{ IMPORTS                            }....................
from beartype._check.metadata.metacheck import BeartypeCheckMeta
from beartype._check.metadata.metadecor import BeartypeDecorMeta
from beartype._check.signature.sigmake import make_func_signature
from beartype._data.code.datacodefunc import (
    CODE_RETURN_UNCHECKED_format,
    CODE_SIGNATURE,
)
from beartype._data.code.datacodename import (
    ARG_NAME_CHECK_META,
    ARG_NAME_FUNC,
)
from beartype._decor._nontype._wrap._wrapargs import (
    code_check_args as _code_check_args)
from beartype._decor._nontype._wrap._wrapreturn import (
    code_check_return as _code_check_return)

# ....................{ GENERATORS                         }....................
def generate_code(decor_meta: BeartypeDecorMeta) -> str:
    '''
    Generate a Python code snippet dynamically defining the wrapper function
    type-checking the passed decorated callable.

    This high-level function implements this decorator's core type-checking,
    converting all unignorable PEP-compliant type hints annotating this
    callable into pure-Python code type-checking the corresponding parameters
    and return values of each call to this callable.

    Parameters
    ----------
    decor_meta : BeartypeDecorMeta
        Decorated callable to be type-checked.

    Returns
    -------
    str
        Generated function wrapper code. Specifically, either:

        * If the decorated callable requires *no* type-checking (e.g., due to
          all type hints annotating this callable being ignorable), the empty
          string. Note this edge case is distinct from a related edge case at
          the head of the :func:`beartype.beartype` decorator reducing to a noop
          for unannotated callables. By compare, this boolean is ``True`` only
          for callables annotated with **ignorable type hints** (i.e.,
          :class:`object`, :class:`beartype.cave.AnyType`, :class:`typing.Any`):
          e.g.,

          .. code-block:: python

             >>> from beartype.cave import AnyType
             >>> from typing import Any
             >>> def muh_func(
             ...     muh_param1: AnyType, muh_param2: object) -> Any: pass
             >>> muh_func is beartype(muh_func)
             True

        * Else, a code snippet defining the wrapper function type-checking the
          decorated callable, including (in order):

          * A signature declaring this wrapper, accepting both beartype-agnostic
            and -specific parameters. The latter include:

            * A private ``__beartype_func`` parameter initialized to the
              decorated callable. In theory, this callable should be accessible
              as a closure-style local in this wrapper. For unknown reasons
              (presumably, a subtle bug in the exec() builtin), this is *not*
              the case. Instead, a closure-style local must be simulated by
              passing this callable at function definition time as the default
              value of an arbitrary parameter. To ensure this default is *not*
              overwritten by a function accepting a parameter of the same name,
              this unlikely edge case is guarded against elsewhere.

          * Statements type checking parameters passed to the decorated
            callable.
          * A call to the decorated callable.
          * A statement type checking the value returned by the decorated
            callable.

    Raises
    ------
    BeartypeDecorParamNameException
        If the name of any parameter declared on this callable is prefixed by
        the reserved substring ``__bear``.
    BeartypeDecorHintNonpepException
        If any type hint annotating any parameter of this callable is neither:

        * **PEP-compliant** (i.e., :mod:`beartype`-agnostic hint compliant with
          annotation-centric PEPs).
        * **PEP-noncompliant** (i.e., :mod:`beartype`-specific type hint *not*
          compliant with annotation-centric PEPs)).
    _BeartypeUtilMappingException
        If generated code type-checking any pair of parameters and returns
        erroneously declares an optional private beartype-specific parameter of
        the same name with differing default value. Since this should *never*
        happen, a private non-human-readable exception is raised in this case.
    '''

    # ....................{ ARGS                           }....................
    # Python code snippet type-checking all callable parameters if one or more
    # such parameters are annotated with unignorable type hints *OR* the empty
    # string otherwise.
    code_check_params = _code_check_args(decor_meta)

    # ....................{ (RETURN|YIELD)                 }....................
    # Python code snippet type-checking the callable return if this return is
    # annotated with an unignorable type hint *OR* the empty string otherwise.
    code_check_return = _code_check_return(decor_meta)

    # If the callable return requires *NO* type-checking...
    #
    # Note that this branch *CANNOT* be embedded in the prior call to the
    # code_check_return() function, as doing so would prevent us from
    # efficiently reducing to a noop here.
    if not code_check_return:
        # If all callable parameters also require *NO* type-checking, this
        # callable itself requires *NO* type-checking. In this case, return the
        # empty string instructing the parent @beartype decorator to reduce to a
        # noop (i.e., the identity decorator returning this callable as is).
        if not code_check_params:
            return ''
        # Else, one or more callable parameters require type-checking.

        # Python code snippet calling this callable unchecked, returning the
        # value returned by this callable from this wrapper.
        code_check_return = CODE_RETURN_UNCHECKED_format(
            func_call_prefix=decor_meta.func_wrapper_code_call_prefix,
            func_return_prefix=decor_meta.func_wrapper_code_return_prefix,
        )
    # Else, the callable return requires type-checking.

    # ....................{ SCOPE                          }....................
    # Dictionary mapping from the name to value of each attribute referenced in
    # the signature of this wrapper function, localized merely for readability.
    func_scope = decor_meta.func_wrapper_scope

    # Expose private beartype type-checking call metadata (i.e.,
    # "beartype"-specific hidden parameter whose default value is the
    # "BeartypeCheckMeta" dataclass instance encapsulating *ALL* metadata
    # required by each call to this wrapper function) to this wrapper function.
    # Doing so dramatically simplifies calls to the get_func_pith_violation()
    # getter inside the body of this wrapper function by enabling this metadata
    # to be passed as a single unified parameter (rather than individually as
    # multiple distinct parameters).
    func_scope[ARG_NAME_CHECK_META] = BeartypeCheckMeta.make_from_decor_meta(
        decor_meta)

    # Expose the callable currently being decorated to this wrapper function.
    # Technically, doing so is merely an optimization; this callable is also
    # accessible as the "ARG_NAME_CHECK_META.func" instance variable in the body
    # of this wrapper function. Pragmatically, doing so is a trivial
    # optimization that could yield non-trivial benefits (e.g., if this wrapper
    # function is frequently called).
    func_scope[ARG_NAME_FUNC] = decor_meta.func_wrappee

    # ....................{ SIGNATURE                      }....................
    # Python code snippet declaring the signature of this type-checking wrapper
    # function, deferred for efficiency until *AFTER* confirming that a wrapper
    # function is even required.
    code_signature = make_func_signature(
        func_name=decor_meta.func_wrapper_name,
        func_scope=func_scope,
        code_signature_format=CODE_SIGNATURE,
        code_signature_prefix=decor_meta.func_wrapper_code_signature_prefix,
        conf=decor_meta.conf,
    )

    # ....................{ TYPE-CHECK                     }....................
    # Return Python code defining the wrapper type-checking this callable.
    #
    # While there exist numerous alternatives to string formatting (e.g.,
    # appending to a list or bytearray before joining the items of that
    # iterable into a string), these alternatives are either:
    # * Slower, as in the case of a list (e.g., due to the high up-front cost
    #   of list construction).
    # * Cumbersome, as in the case of a bytearray.
    #
    # Since string concatenation is heavily optimized by the official CPython
    # interpreter, the simplest approach is the most ideal. KISS, bro.
    return f'{code_signature}{code_check_params}{code_check_return}'



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype Inferential Type-hint Engine (BITE) builtin collections type hint
inferrers** (i.e., lower-level functions dynamically inferring subscripted type
hints describing instances of builtin C-based container types).
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import (
    Dict,
    ChainMap,
    Counter,
    Deque,
    FrozenSet,
    KeysView,
    List,
    Optional,
    Set,
    Tuple,
    ValuesView,
)
from beartype._cave._cavefast import (
    DictKeysViewType,
    DictValuesViewType,
)
from beartype._data.typing.datatyping import (
    DictTypeToAny,
)
from beartype._util.cache.utilcachecall import callable_cached
from collections import (
    ChainMap as ChainMapType,
    Counter as CounterType,
    deque,
)

# ....................{ INFERERS                           }....................
def infer_hint_collection_builtin(obj: object, **kwargs) -> Optional[object]:
    '''
    **Builtin collection type hint** (i.e., subscripted C-based type whose
    instances contain one or more reiterable items) recursively validating the
    passed object (including *all* items transitively reachable from this
    object) if this object is an instance of a builtin collection type *or*
    :data:`None` otherwise (i.e., if this object is *not* an instance of a
    builtin collection type).

    This function *cannot* be memoized, due to necessarily accepting the
    ``__beartype_obj_ids_seen__`` parameter unique to each call to the parent
    :func:`beartype.bite.infer_hint` function. Moreover, this function exhibits
    worst-case constant time complexity :math:`O(1)`; memoization is irrelevant.

    Parameters
    ----------
    obj : object
        Object to infer a type hint from.

    All remaining keyword parameters are passed as is to the lower-level
    :func:`beartype.bite.collection.infercollectionitems.infer_hint_collection_items`
    function.

    Returns
    -------
    Optional[object]
        Either:

        * If this object is a builtin collection type, this type subscripted by
          the union of the child type hints validating *all* items of this
          collection (e.g., ``list[str]``).
        * Else, :data:`None`.
    '''

    # Hint to be returned, defaulting to "None" as a fallback.
    hint: object = None

    # Type of this object.
    obj_type = obj.__class__
    # print(f'Inferring possibly builtin collection type {repr(obj_type)}...')

    # Type hint factory creating type hints validating this type if this
    # type is a subclass of a builtin collection type *OR* "None" otherwise.
    hint_factory = _infer_hint_factory_collection_builtin(obj_type)

    # If this type is a subclass of a builtin collection type...
    if hint_factory:
        # print(f'Inferring iterable {repr(obj_type_collections_abc)} subscription...')

        # Avoid circular import dependencies.
        from beartype.bite.collection.infercollectionitems import (
            infer_hint_collection_items)

        # Hint recursively validating this collection (including *ALL* items
        # transitively reachable from this collection), defined by subscripting
        # this collection type by the union of the child type hints validating
        # all items recursively reachable from this collection.
        hint = infer_hint_collection_items(
            obj=obj, hint_factory=hint_factory, **kwargs,)  # type: ignore[arg-type]
    # Else, this type is *NOT* a subclass of a builtin collection type. In this
    # case, fallback to returning "None".

    # Return this hint.
    return hint

# ....................{ PRIVATE ~ inferers                 }....................
@callable_cached
def _infer_hint_factory_collection_builtin(cls: type) -> Optional[object]:
    '''
    **Builtin collection superclass** (i.e., unsubscripted C-based type whose
    instances contain one or more reiterable items) of the passed class if this
    class subclasses a builtin collection type *or* :data:`None` otherwise
    (i.e., if this class does *not* subclass a builtin collection type).

    This function does *not* return a type hint annotating the passed class.
    This function merely returns a subscriptable type hint factory suitable for
    the caller to subscript with a child type hint, which then produces the
    desired type hint annotating the passed class. Why this indirection? In a
    word: efficiency. If this function recursively inferred and returned a full
    type hint, this function would need to guard against infinite recursion by
    accepting the ``__beartype_obj_ids_seen__`` parameter also accepted by the
    parent :func:`beartype.bite.inferhint.infer_hint` function;
    doing so would prevent memoization, substantially reducing efficiency.

    This function is memoized for efficiency.

    Parameters
    ----------
    cls : type
        Class to be introspected.

    Returns
    -------
    Optional[object]
        Either:

        * If this class subclasses a builtin collection type, the type hint
          factory validating that type.
        * Else, :data:`None`.
    '''
    assert isinstance(cls, type), f'{repr(cls)} not type.'

    # ....................{ IS                             }....................
    # Hint factory describing this class if this class is a builtin collection
    # type (e.g., "list", "tuple") *OR* "None" otherwise.
    hint_factory: object = _COLLECTION_BUILTIN_TYPE_TO_HINT_FACTORY_get(cls)

    # If this class is a builtin collection type, return this hint factory.
    if hint_factory:
        # print(f'Inferred builtin collection {repr(cls)} factory {repr(hint_factory)}...')
        return hint_factory
    # Else, this class is *NOT* a builtin collection type. However, this class
    # could still be a proper subclass of a builtin collection type: e.g.,
    #     class MuhList(list): ...
    #
    # Further logic is required to detect which collection type this class
    # subclasses if any.

    # ....................{ SUBCLASS                       }....................
    # Set of all superclasses of the passed class (including the passed class,
    # which is of course its own superclass *AND* subclass, because set theory
    # just goes hard like that).
    #
    # Note that this includes the irrelevant root "object" superclass, which is
    # *NOT* a builtin collection type and is thus guaranteed to *NEVER* be
    # matched below. Although that superclass *COULD* be trivially sliced off
    # (e.g., with an assignment resembling "classes = set(cls.__mro__[:-1])"),
    # doing so only uselessly consumes more time than it saves. So it goes.
    types = set(cls.__mro__)

    # Intersection of the set of all superclasses of this class with the set of
    # all builtin collection types.
    types_collection_builtin = types & _COLLECTION_BUILTIN_TYPES

    # Return either...
    return (
        # If this intersection is non-empty, this class subclasses one or more
        # builtin collection types. In this case, reduce to this class as is.
        # Since *ALL* builtin containers types are PEP 585-compliant
        # subscriptable type hint factories under Python >= 3.9 (e.g.,
        # "list[str]") and since this class subclasses a builtin container type,
        # this subclass is necessarily also implicitly a PEP 585-compliant
        # subscriptable type hint factory.
        cls
        if types_collection_builtin else
        # Else, this intersection is empty, implying this class does *NOT*
        # subclass a builtin collection type. In this case, reduce to a noop.
        None
    )

# ....................{ PRIVATE ~ mappings                 }....................
#FIXME: Also add:
#* "ItemsView". Actually, "ItemsView" support will probably have to be
#  implemented manually. "ItemsViews" are really just iterables over 2-tuples.

# Note that key-value pairs are intentionally defined in decreasing order of
# real-world commonality to reduce time costs in the average case for downstream
# functions leveraging this dictionary (e.g., the
# beartype.bite.collection.infercollectionbuiltin.infer_hint_collection_builtin()
# function).
_COLLECTION_BUILTIN_TYPE_TO_HINT_FACTORY: DictTypeToAny = {
    # Single-argument builtin reiterable types (i.e., C-based collections whose
    # hint factories are subscriptable by only a single child type hint).
    tuple: Tuple,
    list: List,
    frozenset: FrozenSet,
    set: Set,
    deque: Deque,
    DictKeysViewType: KeysView,
    DictValuesViewType: ValuesView,

    # Dual-argument builtin mapping types (i.e., C-based collections whose hint
    # factories are subscriptable by a pair of child key and value type hints).
    dict: Dict,
    ChainMapType: ChainMap,
    CounterType: Counter,
}
'''
Dictionary mapping from each **builtin collection type** (i.e., C-based type
satisfying the :class:`collections.abc.Collection` protocol described by a
corresponding type hint factory) to that factory.
'''


_COLLECTION_BUILTIN_TYPE_TO_HINT_FACTORY_get = (
    _COLLECTION_BUILTIN_TYPE_TO_HINT_FACTORY.get)
'''
:meth:`._COLLECTION_BUILTIN_TYPE_TO_HINT_FACTORY.get` method, globalized for
negligible efficiency gains.
'''


_COLLECTION_BUILTIN_TYPE_TO_HINT_FACTORY_ITEMS = (
    _COLLECTION_BUILTIN_TYPE_TO_HINT_FACTORY.items())
'''
Iterable of all builtin collection types (e.g., :class:`list`) and the type hint
factories describing those types (e.g., :obj:`typing.List`).
'''

# ....................{ PRIVATE ~ sets                     }....................
_COLLECTION_BUILTIN_TYPES = _COLLECTION_BUILTIN_TYPE_TO_HINT_FACTORY.keys()
'''
Set of all **builtin collection types** (i.e., C-based types satisfying the
:class:`collections.abc.Collection` protocol).
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **executable submodule beartype import hook package initialization
submodule** (i.e., data module mimicking real-world usage of the
:func:`beartype.claw.beartype_this_package` import hook from a top-level
third-party package submodule ``{some_package}.__init__`).
'''

# ....................{ IMPORTS                            }....................
from beartype import BeartypeConf
from beartype.claw import beartype_this_package  # <-- boilerplate for victory
# print(f'__init__.__name__: {__name__}')
# print(f'__init__.__package__: {__package__}')

# ....................{ HOOKS                              }....................
# Install a beartype import hook for the current data subpackage directly
# containing this submodule, configured by the default beartype configuration.
beartype_this_package()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`695` **type alias data submodule.**

This submodule defines :pep:`695`-compliant type aliases to be exercised
throughout unit tests exercising these aliases. For safety, these aliases are
intentionally isolated from those unit tests.

Motivation
----------
Generic type alias support is unavoidably implemented in an *extremely* fragile
manner throughout the :mod:`beartype` codebase. Recursively "bubbling up" the
concrete child type hints subscripting a generic type alias into the lower-level
target type hint aliased by that generic type alias is so non-trivial that it
necessitates implementing manual support for "bubbling up" across both all
hint-specific dynamic code generators *and* all hint-specific dynamic exception
raisers: which is to say, the *entire* :mod:`beartype._check` subpackage.

Generic type alias support is so widely distributed that the *only* valid
means of testing this support is to exhaustively define here and subsequently
use below at least one generic type alias for *each* other PEP standard
supported by :mod:`beartype`. Non-trivial, thy name is :pep:`695`.

Caveats
-------
**This submodule requires the active Python interpreter to target at least
Python 3.12.0.** If this is *not* the case, importing this submodule raises an
:exc:`SyntaxError` exception.
'''

# ..................{ IMPORTS                                }..................
from beartype.typing import (
    Annotated,
    Union,
)
from beartype.vale import Is
from beartype_test.a00_unit.data.pep.generic.data_pep585generic import (
    Pep585IterableTContainerT,
    Pep585IterableTupleSTContainerTupleST,
)

# ..................{ SIMPLE                                 }..................
type AliasSimple = int | list[str]
'''
Simple type alias aliasing a standard type hint containing *no* syntax or
semantics unique to PEP 695-compliant type aliases (e.g., *no* forward
references, recursion, or type variables).
'''

# ..................{ PEP (484|604)                          }..................
type AliasPep484604[T] = (
    # PEP-noncompliant type.
    float |
    # PEP 604-compliant new union of two or more generic type hints
    # parametrized by the same type variable subscripting this alias.
    list[T] | set[T] |
    # PEP 484-compliant old union of (...waitforit) two or more generic type
    # hints parametrized by the same type variable subscripting this alias.
    Union[dict[T, T] | frozenset[T]]
)
'''
Type alias aliasing a :pep:`604`-compliant new union of:

* A PEP-noncompliant type.
* Two or more generic type hints parametrized by the same type variable
  subscripting this alias.
* A :pep:`484`-compliant old union of two or more generic type hints
  parametrized by the same type variable subscripting this alias.

This alias thus tests that dynamic code generation for both :pep:`484`- and
:pep:`604`-compliant unions preserves :pep:`484`-compliant type variable
mappings while also flattening directly nested unions into the top-level union.
'''

# ..................{ PEP (484|604) ~ depth ~ unparametrized }..................
type AliasPep484604Depth1 = _AliasPep484604Depth2 | float
'''
Type alias aliasing a :pep:`604`-compliant new union of a PEP-noncompliant type
and another type alias that has yet to be defined aliasing a
:pep:`484`-compliant old union of another PEP-noncompliant type and another type
alias aliasing (...waitforit) another PEP-noncompliant type.

This alias thus tests that dynamic code generation for both :pep:`484`- and
:pep:`604`-compliant unions recursively unflattens deeply nested unions into the
top-level union.
'''


type _AliasPep484604Depth2 = Union[str, _AliasPep484604Depth3]
'''
Type alias aliasing a :pep:`484`-compliant old union of a PEP-noncompliant type
and another type alias aliasing another PEP-noncompliant type that has yet to be
defined.
'''


type _AliasPep484604Depth3 = int
'''
Type alias aliasing a PEP-noncompliant type.
'''

# ..................{ PEP (484|604) ~ depth ~ parametrized   }..................
type AliasPep484604Depth1T[T] = _AliasPep484604Depth2T[T] | float
'''
Type alias aliasing a :pep:`604`-compliant new union of a PEP-noncompliant type
and another type alias subscripted by the same type variable parametrizing this
alias aliasing a :pep:`484`-compliant old union of another PEP-noncompliant type
and another type alias subscripted by the same type variable parametrizing this
alias aliasing that type parameter.

This alias thus tests that dynamic code generation for both :pep:`484`- and
:pep:`604`-compliant unions recursively unflattens deeply nested unions into the
top-level union.
'''


type _AliasPep484604Depth2T[T] = Union[str, _AliasPep484604Depth3T[T]]
'''
Type alias aliasing a :pep:`484`-compliant old union of a PEP-noncompliant type
and another type alias subscripted by the same type variable parametrizing this
alias aliasing another PEP-noncompliant type .
'''


type _AliasPep484604Depth3T[T] = T
'''
**Identity type alias** isomorphically aliasing the type parameter parametrizing
this alias.
'''

# ..................{ PEP 585                                }..................
type AliasPep585DictST[S, T] = dict[S, T]
'''
Type alias aliasing a :pep:`585`-compliant dictionary subscripted by the same
type variables as those parametrizing this alias.
'''


type AliasPep585TupleFixedST[S, T] = tuple[S, T]
'''
Type alias aliasing a :pep:`585`-compliant fixed-length tuple type hint
subscripted by the same type variables as those parametrizing this alias.
'''


type AliasPep585TypeT[T] = type[T]
'''
Type alias aliasing a :pep:`585`-compliant subclass type hint subscripted by the
same type variable as that parametrizing this alias.
'''

# ..................{ PEP 585 ~ generic                      }..................
type AliasPep585IterableTupleSTContainerTupleST[S, T] = (
    Pep585IterableTupleSTContainerTupleST[S, T])
'''
Type alias aliasing a :pep:`585`-compliant generic subscripted by the same type
variables as those parametrizing this alias.
'''


type AliasPep585IterableTContainerT[T] = T | Pep585IterableTContainerT[T] | None
'''
Type alias aliasing a :pep:`604`-compliant union over a type variable, a
:pep:`585`-compliant generic subscripted by that same type variable, and
:data:`None`.

This alias intentionally exercises an edge case that previously induced infinite
recursion. Badness lurks below.
'''

# ..................{ PEP 593                                }..................
type AliasPep593[T] = Annotated[T | bytes, Is[lambda obj: bool(obj)]]
'''
Type alias aliasing a PEP 593-compliant metahint subscripted by the same
type variable as that parametrizing this alias.
'''

# ..................{ PEP 484                                }..................
type AliasPep593[T] = Annotated[T | bytes, Is[lambda obj: bool(obj)]]
'''
Type alias aliasing a PEP 593-compliant metahint subscripted by the same
type variable as that parametrizing this alias.
'''


# ..................{ RECURSIVE                              }..................
type AliasRecursiveIdentityT[T] = AliasRecursiveIdentityT[T]
'''
**Identity recursive type alias** (i.e., alias recursively aliasing itself).

Although technically valid, this alias is semantically meaningless and thus
reducible to the ignorable singleton.
'''

# ..................{ RECURSIVE ~ pep (484|604) : direct     }..................
# Type aliases aliasing unions directly containing the same aliases recursively.

type AliasPep484604Recursive1T[T] = T | AliasPep484604Recursive1T[T]
'''
Type alias aliasing a :pep:`604`-compliant recursive new union of itself and a
:pep:`484`-compliant type variable.

As with the identity recursive type alias, union members that recursively refer
to the same alias are meaningless and thus reducible to the ignorable singleton.
Ergo, this alias semantically reduces to the equivalent simpler alias:

.. code-block:: python

   type AliasPep484604Recursive1T[T] = T
'''


type AliasPep484604Recursive2T[T] = T | list[AliasPep484604Recursive2T[T]]
'''
Type alias aliasing a :pep:`604`-compliant recursive new union of:

* A :pep:`484`-compliant type variable.
* A generic type hint recursively subscripted by this alias.
'''

# ..................{ RECURSIVE ~ pep (484|604) : indirect   }..................
# Chains of two or more type aliases aliasing unions indirectly containing one
# of those same aliases recursively.

type AliasPep484604RecursiveTopT[T] = (
    AliasPep484604RecursiveBottomT[T] | int | bool | float | bytes)
'''
Type alias aliasing a :pep:`604`-compliant recursive new union of:

* Another type alias itself aliasing a :pep:`604`-compliant recursive new union
  of this same alias, inducing recursion via an indirect chain of aliases.
* Arbitrary types.
'''


type AliasPep484604RecursiveBottomT[T] = (
    set[AliasPep484604RecursiveTopT[T]] |
    dict[AliasPep484604RecursiveTopT[T], T] |
    list[AliasPep484604RecursiveBottomT[T]]
)
'''
Type alias aliasing a :pep:`604`-compliant recursive new union of two or more
generic type hints recursively subscripted by:

* Another type alias itself aliasing a :pep:`604`-compliant recursive new union
  of this same alias, inducing recursion via an indirect chain of aliases.
* This same alias.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

"""
**Beartype ANSI utility unit tests.**

This submodule unit tests the public API of the private
:mod:`beartype._util.text.utiltextansi` submodule.
"""


# ....................{ IMPORTS                            }....................
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ tester                     }....................
def test_is_text_ansi() -> None:
    '''
    Test the :func:`beartype._util.text.utiltextansi.is_str_ansi` tester.
    '''

    # Defer test-specific imports.
    from beartype._util.text.utiltextansi import is_str_ansi

    # Assert that this tester returns false for a string containing *NO* ANSI.
    assert is_str_ansi('The sea-blooms and the oozy woods which wear') is False

    # Assert that this tester returns true for a string containing ANSI.
    assert is_str_ansi('The sapless foliage of the ocean, know\033[92m') is (
        True)

# ....................{ TESTS ~ stripper                   }....................
def test_strip_str_ansi() -> None:
    '''
    Test the :func:`beartype._util.text.utiltextansi.strip_str_ansi` stripper.
    '''

    # Defer test-specific imports.
    from beartype._util.text.utiltextansi import strip_str_ansi

    # String containing *NO* ANSI.
    THY_VOICE = 'and suddenly grow gray with fear,'

    # Assert that this stripper preserves strings containing *NO* ANSI as is.
    assert strip_str_ansi(THY_VOICE) == THY_VOICE

    # Assert that this stripper strips *ALL* ANSI from strings containing ANSI.
    assert strip_str_ansi(f'\033[31m{THY_VOICE}\033[92m') == THY_VOICE



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`557` **dataclasses data submodule.**

This submodule exercises support for **dataclasses** (i.e., :pep:`557`-compliant
:func:`dataclasses.dataclass`-decorated classes) implemented in the
:func:`beartype.beartype` decorator. To do so, this submodule defines
dataclasses configured with varying combinations of keyword parameters.
'''

# ....................{ IMPORTS                            }....................
from dataclasses import dataclass

# ....................{ CLASSES                            }....................
@dataclass
class DataclassDefault(object):
    '''
    Arbitrary dataclass declared with default :func:`.dataclass` parameters.
    '''

    pass


@dataclass(frozen=True)
class DataclassFrozen(object):
    '''
    Arbitrary dataclass declared with non-default :func:`.dataclass` parameters
    configuring a **frozen** (i.e., immutable) dataclass.
    '''

    pass
